'''
   FigAxlineC

     A figplot which a user can edit value by mouse
'''
import numpy as np
import wx
import os
import ifigure
from ifigure.mto.fig_axline import FigAxline
from ifigure.utils.geom import transform_point
from ifigure.utils.cbook import ProcessKeywords, LoadImageFile
from ifigure.widgets.undo_redo_history import GlobalHistory, UndoRedoFigobjMethod
from ifigure.utils.edit_list import DialogEditList
from ifigure.mto.fig_control import FigControl


def _copy(x):
    if x is None:
        return
    return x.copy()


class FigAxlineC(FigAxline, FigControl):
    def __new__(cls, *args, **kargs):
        draggable, kargs = ProcessKeywords(kargs, 'draggable', True)
        dragmode, kargs = ProcessKeywords(kargs, 'dragmodee', 'independent')
        enabled_point, kargs = ProcessKeywords(kargs, 'enabled_point', 'all')
        obj = FigAxline.__new__(cls, *args, **kargs)
        if obj is not None:
            obj.setvar('draggable', draggable)
            obj.setvar('dragmode', dragmode)
            if not obj.hasvar('enabled_point'):
                obj.setvar('enabled_point', enabled_point)
        return obj

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = LoadImageFile(path, 'plotc.png')
        return [idx1]

    @classmethod
    def get_namebase(self):
        return 'axlinec'

    def __init__(self, *args, **kargs):
        FigAxline.__init__(self, *args, **kargs)
        FigControl.__init__(self)
        self._figc_hit = -1
        self._enable_a = None

    @classmethod
    def attr_in_file(self):
        return ['enabled_point']+super(FigAxlineC, self).attr_in_file()

    def args2var(self):
        val = FigAxline.args2var(self)
        if self.getvar('enabled_point') == 'all':
            self.setp('enabled_point',
                      [True]*(self.getp('x').size + self.getp('y').size))
        return val

    def generate_artist(self):
        FigAxline.generate_artist(self)
        idx = np.where([not x for x in self.getp('enabled_point')])[0]
        if self._enable_a is not None:
            self._enable_a.remove()
            self._enable_a = None
        if idx.size == 0:
            return
        self._enable_a = []  # it is actually for disabled points
        x, y = self._eval_xy()
        container = self.get_container()
        for k in idx:
            lw = self._artists[k].get_linewidth()
            if k < x.size:
                self._enable_a.append(container.axvline(x[k],
                                                        color='white',
                                                        linestyle='--',
                                                        linewidth=lw))
            else:
                self._enable_a.append(container.axhline(y[k-x.size],
                                                        color='white',
                                                        linestyle='--',
                                                        linewidth=lw))

    def del_artist(self, artist=None, delall=False):
        if self._enable_a is not None:
            for a in self._enable_a:
                a.remove()
            self._enable_a = None
        return FigAxline.del_artist(self, artist=artist, delall=delall)

    def canvas_menu(self):
        m = [('Add Point', self.onAddPoint, None), ]
        if self._figc_hit != -1:
            if self.getp('enabled_point')[self._figc_hit]:
                m = m + [('Delete Point', self.onDelPoint, None),
                         ('Edit Point', self.onEditPoint, None),
                         ('Disable Point', self.onDisablePoint, None), ]
            else:
                m = m + [('Delete Point', self.onDelPoint, None),
                         ('Edit Point', self.onEditPoint, None),
                         ('Enable Point', self.onEnablePoint, None), ]
        return m + FigAxline.canvas_menu(self)

    def picker_a(self, a, evt):
        hit, extra = FigAxline.picker_a(self, a, evt)
        if hit:
            self._figc_hit = self._artists.index(extra['child_artist'])
            self._drag_backup = (a.get_xdata(), a.get_ydata())
        else:
            self._figc_hit = -1
        return hit, extra

    def drag(self, a, evt):
        if self.getvar('dragmode') == 'independent':
            if evt.inaxes is None:
                return 0
            if evt.xdata is None:
                return

            axes = a.axes
            x, y = self._eval_xy()
            if self._figc_hit < x.size:
                #                self.setp('x')[self._figc_hit] = evt.xdata
                self._artists[self._figc_hit].set_xdata([evt.xdata, evt.xdata])
            else:
                #                self.setp('y')[self._figc_hit - x.size] = evt.ydata
                self._artists[self._figc_hit].set_ydata([evt.ydata, evt.ydata])
        else:
            return FigAxline.drag(self, a, evt)

    def dragdone(self, a, evt):
        if self.getvar('dragmode') == 'independent':
            axes = a.axes
            x, y = self._eval_xy()
            if self._figc_hit < x.size:
                x = self.getp('x').copy()
                x[self._figc_hit] = evt.xdata
                y = self.getp('y').copy()
            else:
                x = self.getp('x').copy()
                y = self.getp('y').copy()
                y[self._figc_hit - x.size] = evt.ydata

            action = UndoRedoFigobjMethod(self._artists[0],
                                          'data',
                                          (x, y))
            window = evt.guiEvent.GetEventObject().GetTopLevelParent()
            GlobalHistory().get_history(window).make_entry([action])
            return 1
        else:
            return FigAxline.dragdone(self, a, evt)

    def do_update_artist(self):
        if not self.isempty():
            try:
                a1 = self._artists[self._figc_hit]
            except:
                a1 = self._artists[0]
            self.del_artist(delall=True)
        if self.isempty() and not self._suppress:
            self.generate_artist()
            try:
                a2 = self._artists[self._figc_hit]
            except:
                a2 = self._artists[0]
            if a1 != a2:
                ifigure.events.SendReplaceEvent(self, a1=a1, a2=a2)

    def onAddPoint(self, evt):
        x = self.getp('x').copy()
        y = self.getp('y').copy()
        v = self.getp('enabled_point')[:]
        if self._figc_hit < x.size:
            x = np.insert(x, self._figc_hit + 1, self._drag_backup[0][0])
            v.insert(self._figc_hit + 1, True)
        else:
            y = np.insert(y, self._figc_hit + 1 - x.size,
                          self._drag_backup[1][0])
            v.insert(self._figc_hit + 1, True)
        actions = [UndoRedoFigobjMethod(self._artists[0],
                                        'data',
                                        (x, y)),
                   UndoRedoFigobjMethod(self._artists[0],
                                        'enabled_point', v), ]
        window = evt.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(actions, menu_name='add point')
        return 1

    def onDelPoint(self, evt):
        x = self.getp('x').copy()
        y = self.getp('y').copy()
        v = self.getp('enabled_point')[:]
        if self._figc_hit < x.size:
            x = np.delete(x, self._figc_hit)
            del v[self._figc_hit]
        else:
            y = np.delete(y, self._figc_hit - x.size)
            del v[self._figc_hit]

        actions = [UndoRedoFigobjMethod(self._artists[0],
                                        'data',
                                        (x, y)),
                   UndoRedoFigobjMethod(self._artists[0],
                                        'enabled_point', v), ]
        window = evt.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(actions, menu_name='delete point')
        return 1

    def onEditPoint(self, evt):
        x, y = self._eval_xy()
        if self._figc_hit < x.size:
            l = [['x', str(x[self._figc_hit]), 0, None], ]
        else:
            l = [['y', str(y[self._figc_hit-x.size]), 0, None], ]
        window = evt.GetEventObject().GetTopLevelParent()
        value = DialogEditList(l, modal=True,
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
                               tip=None,
                               parent=window)
        if value[0]:
            v = value[1]
        else:
            return
        x = self.getp('x').copy()
        y = self.getp('y').copy()
        if self._figc_hit < x.size:
            x[self._figc_hit] = float(v[0])
        else:
            y[self._figc_hit - x.size] = v[0]
        action = UndoRedoFigobjMethod(self._artists[0],
                                      'data',
                                      (x, y))
        window = evt.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry([action])
        return 1

    def onEnablePoint(self, evt):
        v = self.getp('enabled_point')[:]
        v[self._figc_hit] = True
        x = self.getp('x').copy()
        y = self.getp('y').copy()

        actions = [UndoRedoFigobjMethod(self._artists[0],
                                        'enabled_point', v),
                   UndoRedoFigobjMethod(self._artists[0],
                                        'data',
                                        (x, y))]
        window = evt.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(actions)
        return 1

    def onDisablePoint(self, evt):
        v = self.getp('enabled_point')[:]
        v[self._figc_hit] = False
        x = self.getp('x').copy()
        y = self.getp('y').copy()

        actions = [UndoRedoFigobjMethod(self._artists[0],
                                        'enabled_point', v),
                   UndoRedoFigobjMethod(self._artists[0],
                                        'data',
                                        (x, y))]

        window = evt.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(actions)
        return 1

    def set_enabled_point(self, value, a):
        self.setp('enabled_point', value)
        self.set_update_artist_request()
        self.get_figaxes().set_bmp_update(False)

    def get_enabled_point(self, a):
        return self.getp('enabled_point')

    def set_data(self, value, a):
        self.setp('x', value[0])
        self.setp('y', value[1])
        self._data_extent = None
        self.set_update_artist_request()
        self.call_control_changed_callback()

    def get_data(self, a=None):
        return (self.getp('x').copy(), self.getp('y').copy(), self.getp('enabled_point')[:])

    def save_data2(self, data=None):
        var = {'enabled_point': self.getp('enabled_point')}
        data['FigAxlineC'] = (1, var)
        data = super(FigAxlineC, self).save_data2(data)
        return data

    def load_data2(self, data):
        super(FigAxlineC, self).load_data2(data)
        if not 'FigAxlineC' in data:
            self.setp('enabled_point',
                      [True]*(self.getp('x').size + self.getp('y').size))
        else:
            d = data['FigAxlineC'][1]
            self.setp('enabled_point', d['enabled_point'])



from __future__ import print_function

'''
   ifigure.interactive

   these are commands which can be used in piScope
   interactive shell (command input)

   many function calls are actually redicted to
   methods of bookviewer with additional keywords
   from _hold and _update. In that case, redirect_
   to_aviewer (decorator) is used to avoid repeating
   the same code. However, methods in bookviewer are
   programmed to use __doc__ in this file.

'''

import matplotlib.mlab as mlab
from functools import wraps
import logging
import ifigure
import wx
import weakref
import os
import numpy as np
from ifigure.utils.triangulation_wrapper import delaunay
#from ifigure.utils.cbook import isiterable, isndarray, message

_hold = False
_update = True
_lastisec = 0

aviewer = None


def get_topwindow():
    return wx.GetApp().GetTopWindow()


def set_aviewer(viewer):
    # called from ifigure_app.aviewer
    from ifigure.widgets.book_viewer import BookViewer
    if (viewer == wx.GetApp().TopWindow or
            isinstance(viewer, BookViewer)):
        globals()['aviewer'] = viewer
    elif (globals()['aviewer'] is not None):
        try:
            if (globals()['aviewer'].book is None):
                globals()['aviewer'] = None
        except:
            globals()['aviewer'] = None
    if viewer is None:
        if globals()['aviewer'] in wx.GetApp().TopWindow.viewers:
            return
        globals()['aviewer'] = None


def check_aviewer(func):
    @wraps(func)
    def checker(*args, **kargs):
        def func2(*args, **kargs):
            #              return figure()
            from ifigure.utils.cbook import message
            message("*** No current viewer (no plot) ***")
            return None

        # check if aviewer in still right object
        try:
            book = aviewer.book
        except:
            return func2(*args, **kargs)

        if aviewer.book is None:
            globals()['aviewer'] = None
            return func2(*args, **kargs)
        else:
            return func(*args, **kargs)

    return checker


def redirect_to_aviewer(func):
    @wraps(func)
    def checker(*args, **kargs):
        if aviewer is None:
            figure()
            wx.GetApp().Yield()  # yield let wx to process event including
            # project_tree_widget update
        m = getattr(aviewer, func.__name__)
        kargs['hold'] = _hold
        kargs['update'] = _update
        ret = m(*args, **kargs)
        aviewer.Raise()
        return ret
    return checker


def redirect_to_aviewer_3D(func):
    @wraps(func)
    def checker(*args, **kargs):
        if aviewer is None:
            figure()
            wx.GetApp().Yield()  # yield let wx to process event including
            # project_tree_widget update
        m = getattr(aviewer, func.__name__)
        kargs['hold'] = _hold
        kargs['update'] = _update
        aviewer.threed('on')
        ret = m(*args, **kargs)
        aviewer.Raise()
        return ret
    return checker


def redirect_to_aviewer_hold(func):
    @wraps(func)
    def checker(*args, **kargs):
        def func2(*args, **kargs):
            #              from ifigure.utils.cbook import message
            return figure()
#              message("*** No current viewer (no plot) ***")
#              return None
        if aviewer is None:
            figure()
#           return func2(*args, **kargs)
#        else:

        m = getattr(aviewer, func.__name__)
        kargs['hold'] = True
        kargs['update'] = _update
        ret = m(*args, **kargs)
        aviewer.Raise()
        return ret
    return checker


@redirect_to_aviewer
def showpage(ipage):
    '''
    show ipage
    '''
    pass


@redirect_to_aviewer
def cla(reset_color_cycle=True):
    '''
    clear current axis
    '''
    pass


@redirect_to_aviewer
def cls():
    '''
    cls() is the same as clf()
    isec is moved to 0
    '''
    pass


@redirect_to_aviewer
def clf():
    '''
    clear current page. page annotations are not
    deleted
    isec is moved to 0
    '''
    pass


@redirect_to_aviewer
def nsec(*args, **kargs):
    '''
    nsec is the same as subplot
    see subplot help (type 'subplot(' to show help)
    '''
    pass


@redirect_to_aviewer
def nsection(*args, **kargs):
    '''
    nsection is the same as subplot
    see subplot help (type 'subplot(' to show help)
    '''
    pass


@redirect_to_aviewer
def subplot(*args, **kargs):
    '''
    set page section format.
       subplot(3)     3 rows
       subplot(1, 3)     3 columns
       subplot(2, 3)  2x3
       subplot(2, 3, (0,1)) 2x3 and (0,1) merged
       subplot(2, 3, (0,1), (2, 3)) 2x3 and (0,1), (2, 3) merged

       'sort' = 'col' or 'column' or 'c' : sort result in column
       'sort' = 'row' or 'r' :             sort result in row

       dx and dy are optional arguments to determine the
       width and height of each column and row
       if these are used, the number of dx and dy should be
       ncol-1, nrow-1, respectively

       example: subplot(2,2, (0,1), dx=0.4)

    '''
    pass


@redirect_to_aviewer
def isec(i=None):
    '''
    isce/isection control current axes.
    if i is give, it sets current axes and returns ax
    otherwize it returns current ax
    '''
    pass


@redirect_to_aviewer
def isection(i=None):
    '''
    isce/isection control current axes.
    if i is give, it sets current axes and returns ax
    otherwize it returns rrent ax
    '''
    pass


@redirect_to_aviewer
def addpage(num=1, before=False):
    '''
    add a page to current book
    '''
    pass


@redirect_to_aviewer
def delpage():
    '''
    delete current page
    '''


@redirect_to_aviewer
def suptitle(txt, size=None, color=None):
    '''
    set page  title
    '''
    pass


@redirect_to_aviewer
def title(txt, size=None, color=None):
    '''
    set section  title
    '''
    pass


@redirect_to_aviewer
def xlabel(txt, name='x', size=None, color=None):
    '''
    set xaxis label
       xlabel(txt)
       xlabel(txt, name = 'x2')
       xlabel(txt, size=10, color='red')
    '''
    pass


@redirect_to_aviewer
def xtitle(txt, name='x', size=None, color=None):
    '''
    set xaxis label
       xtitle(txt)
       xtitle(txt, name = 'x2')
       xtitle(txt, size=10, color='red')
    '''
    pass


@redirect_to_aviewer
def ylabel(txt, name='y', size=None, color=None):
    '''
    set yaxis label
       ylabel(txt)
       ylabel(txt, name = 'y2')
       ylabel(txt, size=10, color='red')
    '''
    pass


@redirect_to_aviewer
def ytitle(txt, name='y', size=None, color=None):
    '''
    set yaxis label
       ytitle(txt)
       ytitle(txt, name = 'y2')
       ytitle(txt,  size=10, color='red')
    '''
    pass


@redirect_to_aviewer
def zlabel(txt, name='z', size=None, color=None):
    '''
    set zaxis label
       zlabel(txt)
       zlabel(txt, size=10, color='red')
    '''
    pass


@redirect_to_aviewer
def ztitle(*args):
    '''
    set zaxis label
       ztitle(txt)
       ztitle(txt, size=10, color='red')
    '''
    pass


@redirect_to_aviewer
def clabel(txt, name='c'):
    '''
    set caxis label
       clabel(txt)
       clabel(txt, 'c2')
    '''
    pass


@redirect_to_aviewer
def ctitle(*args):
    '''
    set caxis label
       ctitle(txt)
       ctitle(txt, 'c2')
    '''
    pass


@redirect_to_aviewer_hold
def xlog(value=True, base=None):
    '''
    set xlog
    xlog()
    xlog(False)
    '''
    pass


@redirect_to_aviewer_hold
def ylog(value=True, base=None):
    '''
    set ylog
    ylog()
    ylog(False)
    '''
    pass


@redirect_to_aviewer_hold
def clog(value=True, base=None):
    '''
    set ylog
    clog()
    clog(False)
    '''
    pass


@redirect_to_aviewer_hold
def zlog(value=True, base=None):
    '''
    set zlog
    zlog()
    zlog(False)
    '''
    pass


@redirect_to_aviewer_hold
def xsymlog(base=None, linthresh=None, linscale=None, name='x'):
    '''
    set symlog in x

    [x,y, z, c]symlog(base = None, linthresh = None, linscale = None, name  = 'x')
    '''
    pass


@redirect_to_aviewer_hold
def ysymlog(base=None, linthresh=None, linscale=None,  name='y'):
    '''
    set symlog in y
    [x,y, z, c]symlog(base = None, linthresh = None, linscale = None, name  = 'y')
    '''
    pass


@redirect_to_aviewer_hold
def zsymlog(base=None, linthresh=None, linscale=None, name='z'):
    '''
    set symlog in z
    [x,y, z, c]symlog(base = None, linthresh = None, linscale = None,  name  = 'z')
    '''
    pass


@redirect_to_aviewer_hold
def csymlog(base=None, linthresh=None, linscale=None, name='c'):
    '''
    set symlog in c
    [x,y, z, c]symlog(base = None, linthresh = None, linscale = None, name  = 'c')
    '''
    pass


@redirect_to_aviewer_hold
def xlinear(value=True):
    '''
    set xscale linear
    xlinear()
    xlinear(False) # makes log scale
    '''
    pass


@redirect_to_aviewer_hold
def ylinear(value=True):
    '''
    set yscale linear
    ylinear()
    ylinear(False) # makes log scale
    '''
    pass


@redirect_to_aviewer_hold
def clinear(value=True):
    '''
    set cscale linear
    clinear()
    clinear(False) # makes log scale
    '''
    pass


@redirect_to_aviewer_hold
def zlinear(value=True):
    '''
    set zscale linear
    zlinear()
    zlinear(False) # makes log scale
    '''
    pass


@redirect_to_aviewer
def xauto(name='x'):
    '''
    auto scale x
    '''
    pass


@redirect_to_aviewer
def yauto(name='y'):
    '''
    auto scale y
    '''
    pass


@redirect_to_aviewer
def zauto(name='z'):
    '''
    auto scale z
    '''
    pass


@redirect_to_aviewer
def cauto(name='c'):
    '''
    auto scale c
    '''
    pass


@redirect_to_aviewer
def xlim(*range, **kargs):
    '''
    xlim change range of xaxis

    kargs:
        tposition : tick position ('top', 'bottom')
        ticks : tick values
        tcolor : tick color
        color : text color
        size : text size
        ocolor : offset text color
        osize : offset text size

    example) 
        xlim(min, max) , xlim((min, max)), or xlim([min, max])
        xlim([0, 3], size=25, color='red', tcolor='red', ticks=[0,1, 3], tposition='top')
    '''
    pass


@redirect_to_aviewer
def ylim(*range, **kargs):
    '''
    ylim change range of yaxis
    kargs:
        tposition : tick position ('left', 'right')
        ticks : tick values
        tcolor : tick color
        color : text color
        size : text size
        ocolor : offset text color
        osize : offset text size

    example) 
        ylim(min, max) , ylim((min, max)), or ylim([min, max])
        ylim([0, 3], size=25, color='red', tcolor='red', ticks=[0,1, 3], tposition='left')
    '''
    pass


@redirect_to_aviewer
def zlim(*range, **kargs):
    '''
    zlim change range of zaxis

    kargs:
        tposition : tick position ('top', 'bottom')
        ticks : tick values
        tcolor : tick color
        color : text color
        size : text size
        ocolor : offset text color
        osize : offset text size

    example) zlim(min, max) , zlim((min, max)), or zlim([min, max])
    '''
    pass


@redirect_to_aviewer
def clim(*range, **kargs):
    '''
    clim change range of caxis
    example) clim(min, max) , clim((min, max)), or clim([min, max])
    '''
    pass


@redirect_to_aviewer
def twinx():
    '''
    twinx
    '''
    pass


@redirect_to_aviewer
def twiny():
    '''
    twinx
    '''
    pass


@redirect_to_aviewer_hold
def oplot(*args, **kargs):
    '''
    oplot :
        overplot
    see plot for all arguments
    '''
    pass


@redirect_to_aviewer_hold
def oerrorbar(*args, **kargs):
    '''
    oerrrobar:
        overplot errorbar
    see errorbar for all arguments
    '''
    pass


@redirect_to_aviewer
def loglog(*args, **kargs):
    '''
    make loglog plot
    loglog(x, y, s)
    '''
    pass


@redirect_to_aviewer
def semilogy(*args, **kargs):
    '''
    make semilog (yaxis is log scale) plot
    semilogy(x, y, s)
    '''
    pass


@redirect_to_aviewer
def semilogx(*args, **kargs):
    '''
    make semilog (xaxis is log scale) plot
    semilogx(x, y, s)
    '''
    pass


@redirect_to_aviewer
def timetrace(*args, **kargs):
    '''
    timetrace: special plot for time
               it supports decimation
    timetrace(y)
    timetrace(x, y)
    '''
    pass


@redirect_to_aviewer
def plotc(*args, **kargs):
    '''
    plotc creates a line plot similar to plot.
    however, it has extra menus to edit points
    '''
    pass


@redirect_to_aviewer
def errorbarc(*args, **kargs):
    '''
    errorbar creates a line plot similar to errorbar.
    however, it has extra menus to edit points
    '''
    pass


@redirect_to_aviewer
def plot(*args, **kargs):
    """
    plot : xy plot

    plot(y)
    plot(x, y)
    plot(y, s)
`   plot(x, y, s)
    plot(x, y, z)
    plot(x, y, z, cz=True)

    s is a format string. For example 'bo-' means to use blue solid line with
    circle marker. The format string is directly passed to matplotlib.
    See http://piscope.psfc.mit.edu/index.php/Interactive_commands#plot for
    detail

    cz is option to change the color along a line using z.

    When x and y are expression, it evaulate x and y and the answer
    should be 1D data.
    If x and y are given as numbers, following handling
    is done
          x.ndim == 2 and y.ndmi ==1
            x is sliced using the first row and multiple lines
            are generated
          x.ndim == 1 and y.ndmi ==2
            y is sliced using the first row and multiple lines
            are generated
          x.ndim == 2 and y.ndmi ==2
            both x and y are sliced using the first row and multiple lines
            are generated

    see also: errorbar
    """
    pass


@redirect_to_aviewer
def scatter(*args, **kargs):
    """
    scatter plot

    scatter(x, y, s = 20, c = 'b')

    s : scalar or array_like (same length as x, y)
        size in points^2.
    c : color or sequence of color
        if c is a 1D array, it is normalized using c-axis range
        c can also be RGBA values (in which rows are RGB or RGBA)
    """
    pass


@redirect_to_aviewer
def hist(*args, **kargs):
    """
    histgram


    """
    pass


@redirect_to_aviewer
def triplot(*args, **kargs):
    """
    triplot : plot triangles

    triplot(x, y)
    triplot(x, y, mask = mask, ...)
    triplot(tri, x, y)

    """
    pass


@redirect_to_aviewer
def errorbar(*args, **kargs):
    """
    errorbar : xy plot with errorbar

    errorbar(x, y, xerr=xerr, yerr=yerr)

    options for xerr and yerr
        ### assign 0.1 for all points)
        xerr = 0.1
        ### assign different value of error
        xerr = [0.1, 0.2, ....]
        ### assign upper and lower error separately
        xerr = [[0.1, 0.2, ...],[0.4, 0.7...]]

    identical to calling plot with mpl_command = 'errorbar'
    """
    pass


@redirect_to_aviewer
def annotate(*args, **kargs):
    '''
    annotate(s, xy, xytext=None, xycoords='data',
            textcoords='data', arrowprops=None,
            **kargs)
    xycoords, textcoords : 'data', 'figure', 'axes'
      'figure' and 'axes' is from 0 to 1 (fraction)
      other opstions are not supported
    '''
    pass


@redirect_to_aviewer
def ispline(*args, **kargs):
    """
    ispline : xy plot
    ispline(x, y)
    """
    pass


@redirect_to_aviewer
def contour(*args, **kargs):
    """
    contour : contour line plot  (see also contourf)
    contour(z, n)
    contour(x, y, z, n)
    contour(z, v)
    contour(x, y, z, v)

    n: number of levels
    v: a list of contour levels
    """
    pass


@redirect_to_aviewer
def contourf(*args, **kargs):
    """
    contourf : contour fill plot
    contourf(z, n)
    contourf(x, y, z, n)
    contourf(z, v)
    contourf(x, y, z, v)

    n: number of levels
    v: a list of contour levels
    """
    pass


@redirect_to_aviewer
def quiver(*args, **kargs):
    """
    quiver : quiver plot
    for 2D:
       quiver(u, v)
       quiver(u, v, c)
       quiver(x, y, u, v)
       quiver(x, y, u, v, c)

    for 3D:
       quiver(X, Y, Z, U, V, W, **kwargs)

       X, Y, Z:
           The x, y and z coordinates of the arrow locations (default is
           tip of arrow; see *pivot* kwarg)
       U, V, W:
           The x, y and z components of the arrow vectors

    """
    pass


@redirect_to_aviewer
def quiver3d(*args, **kargs):
    '''
    quiver3D is threed('on') + quiver
    quiver3D(x, y, z, u, v, w,  cz = False, cdata = None)

    if cz is True and cdata is None, z is used for color
    '''


@redirect_to_aviewer
def image(*args, **kargs):
    """
    image : show image

    image(z)
    image(x, y, z)
    """
    pass


@redirect_to_aviewer
def specgram(x, NFFT=256,
             Fs=2,
             Fc=0,
             detrend=mlab.detrend_none,
             window=mlab.window_hanning,
             noverlap=128,
             xextent=None,
             pad_to=None,
             sides='default',
             scale_by_freq=None,
             **kwargs):
    '''
    plot spectrogram. Run matplotlib.pyplot.specgram
    and call image using the returnd spectrum.
    keywords are the same as specgram.
    '''
    pass


@redirect_to_aviewer
def spec(*args, **kargs):
    """
    spectram
    spec(t, v)
    spec(v)
    """
    pass


@redirect_to_aviewer
def tripcolor(*args, **kargs):
    """
    tricolor : show image using triangulation

    tripcolor(z)
    tripcolor(x, y, z)
    tripcolor(tri, z)
    tripcolor(tri, x, y, z)

    tri can be evaluated by tri = delaunay(x, y) beforehand
    """
    pass


@redirect_to_aviewer
def tricontour(*args, **kargs):
    """
    tri-contour plot

    tricontour(x, y, z, n)
    tricontour(x, y, z, v)
    tricontour(tri, x, y, z, n)
    tricontour(tri, x, y, z, v)
    tri can be evaluated by tri = delaunay(x, y) beforehand
    """
    pass


@redirect_to_aviewer
def tricontourf(*args, **kargs):
    """
    tri-contour plot with fill mode

    tricontourf(x, y, z, n)
    tricontourf(x, y, z, v)
    tricontourf(tri, x, y, z, n)
    tricontourf(tri, x, y, z, v)
    tri can be evaluated by tri = delaunay(x, y) beforehand
    """
    pass


@redirect_to_aviewer
def axline(*args, **kargs):
    """
    axline : axhline or axvline


    axline(x) or axline([x1,x2,x3...])  : vline
    axline([], y) or axline([], [y1,y2,y3...]) : hline
    axline([x1, x2...],[y1, y2...]) : mixed vline and hline

    (note) lines created by one axline commads shares
           color, marker, alpha, and other attirbute.
    """
    pass


@redirect_to_aviewer
def axlinec(*args, **kargs):
    """
    axlinec: editable axline
    """
    pass


@redirect_to_aviewer
def axspan(*args, **kargs):
    """
    axspan : axhspan or axvspan

    axspan([x1,x2])     : v-span
    axspan([], [y1,y2]) : h-span
    axspan([x1, x2], [y1,y2]) : mixed v-span h-span

    multiple v-span and h-span can be created at once
    axspan([[x1, x2], [x3, x4]...], [[y1,y2], [y3, y4]...])

    (note) artists created by one axspan commands
          shares color, marker, alpha, and other attirbute.
          drag is also applied to all artists.
    """
    pass


@redirect_to_aviewer
def axspanc(*args, **kargs):
    """
    axspanc: a user control version of axspan
             see help(axspan) for the details of argments

             a user can add/drag/remove the patch object and edit
             values from GUI
    """
    pass


@redirect_to_aviewer_hold
def text(*args, **kargs):
    """
    text : add text to current axes

    text(x, y, s) : type string s to (x, y)
    """
    pass


@redirect_to_aviewer_hold
def figtext(*args, **kargs):
    """
    figtext : add text to current figure

    figtext(x, y, s) : type string s to (x, y)
    """
    pass


@redirect_to_aviewer_hold
def arrow(*args, **kargs):
    """
    arrow : add arrow to current axes

    arrow(x1, y1, x2, y2)
    """
    pass


@redirect_to_aviewer_hold
def figarrow(*args, **kargs):
    """
    figarrow : add arrow to current page

    figarrow(x1, y1, x2, y2)
    """
    pass


@redirect_to_aviewer_hold
def legend(*args, **kargs):
    """
    legend: add legend box to current figure

    legend('label1') : legend for a single artist
    legend(['label1', 'label2']) for multiple aritsts
    legend(['label1', 'label2'], axes2 = True) for multiple aritsts
    """
    pass


@redirect_to_aviewer
def fill(*args, **kargs):
    '''
    fill(x, y)
    '''
    pass


@redirect_to_aviewer
def fill_between(*args, **kargs):
    '''
    fill_between(x, y,  y2=[0]*len(x), where=None)
    '''
    pass


@redirect_to_aviewer
def fill_betweenx(*args, **kargs):
    '''
    fill_betweenx(y, x, x2=[0]*len(y), where=None)
    (note): order of x and y is different from MPL?
    '''
    pass


@redirect_to_aviewer
def fill_between_3d(*args, **kargs):
    '''
    fill_between_3d(x1, y1, z1, x2, y, z2, c='b')
    '''
    pass


@redirect_to_aviewer_3D
def surf(*args, **kargs):
    '''
    surf or surface : surface plot in 3D
                      using mplot3d
    surf(x, y, z, **kargs):
    '''
    pass


@redirect_to_aviewer_3D
def surface(x, y, z, **kargs):
    '''
    surf/surface : surface plot in 3D
                      using mplot3d
    surf(x, y, z, **kargs):
    '''
    pass


@redirect_to_aviewer_3D
def revolve(*args, **kargs):
    '''
    revolve r, z : revolve (r, z) data

      keywords to define revolve
         rcenter: [0,0]
         rtheta:  [0, 2*pi]
         raxis:   [0,  1]
         rmesh:   100.
    '''
    pass


@redirect_to_aviewer_3D
def solid(v, **kargs):
    '''
    solid: plot soild volume complsed by triangle/quad

    solid(v, cz=False, cdata=None, **kargs):
    solid(v, idxset, cz=False, cdata=None, **kargs):

    v : 3D array of verteics
        v[ielement, ivertex,  xyz]
    cz : define color data separately
        when cz =true, 3rd dim of v should be four
        v[ielement, ivertex,  xyzc]

    Using idxset, vertices and index set to define the element shape
    is given separately. v[:, xyz] and idxset[ielement, ivertex]
    will be expanded as if v is v[idexset,...]. This allows to reduce
    the number of vertices passed to GPU

    if third dim is 2:
        v[ielement, ivertex,  xy]
        and
        z needs to be given as zvalue keyword argument

    cdata: used with cz  cdata[ielement, ivertex]

    draw_last : draw this artists last on GL canvas, useful for getting
                cleanin line smoothing
    facecolor: use solid facecolor
    edgecolor: use solid edgecolor

    example:

       (indexed array)
       ptx = np.array([[0, 0], [0,1], [1,1], [1,0]])
       box = np.array([[0,1,2,3]])
       figure();solid(ptx, box)
    '''
    pass


@redirect_to_aviewer_3D
def trisurf(v, **kargs):
    '''
    triangle surface plot
    trisurf(z, **kargs):
    trisurf(x, y, z, **kargs):
    trisurf(tri, x, y, z, **kargs):
    trisurf(tri, z, **kargs):
    '''
    pass


@redirect_to_aviewer
def property(obj, name, *args):
    '''
    property set or get property of target object

      property(obj) : return a list of editable property
      property(obj, name) : get an object property
      property(obj, name, value : set an object property
    '''
    pass


@redirect_to_aviewer
def threed(*args):
    '''
    turn on/off three-D axis mode
    '''
    pass


@redirect_to_aviewer
def lighting(**kwargs):
    '''
    set lighting of 3D scene (it affects only artists drawn on OpenGL canvas)
      lighting(ambient = 0.4)  : amibient lighting intensity
      lighting(light   = 0.4)  : lighting source intensity
      lighting(light_direction = (1, 0., 1, 0)) : lighting source direction
      lighting(specular = 1.0) : specular reflection intensity
      lighting(light_color = (1.0, 1, 1)) : light source color
      lighting(wireframe = 0)  : #0 normal mode
                                 #1 wireframe + hidden line elimination
                                 #2 wireframe
    '''
    pass


@redirect_to_aviewer
def _view(*args, **kwargs):
    pass


def view(*args, **kwargs):
    '''
       set 3D view
          view() : return current setting
          view(elev, azim, upvec)
          view('xy')
          view('yx')
          view('xz')
          view('yz')
          view('default')
          view('frustum')
          view('ortho')
          view('updown')
          view('equal')   # equal aspect ratio
          view('auto')    # auto aspect ratio
          view('clip')
          view('noclip')
    '''
    if len(args) == 0 and len(kwargs) == 0:
        v = aviewer
        return v.view()
    else:
        return _view(*args, **kwargs)


@redirect_to_aviewer
def xnames(*args, **kwargs):
    '''
    return list of x axis name of current plot
    '''
    pass


@redirect_to_aviewer
def ynames(*args, **kwargs):
    '''
    return list of y axis name of current plot
    '''
    pass


@redirect_to_aviewer
def znames(*args, **kwargs):
    '''
    return list of z axis name of current plot
    '''
    pass


@redirect_to_aviewer
def cnames(*args, **kwargs):
    '''
    return list of c axis name of current plot
    '''
    pass


@redirect_to_aviewer
def cbar(*args, **kwargs):
    '''
    toggle cbar of current plot
        cbar()
        cbar('c2')  # to specify caxis name

    keywords:
        position : position of color bar (normalized to axis)
        size : size of color bar (normalized to axis)
        direction : h or v (horizontal, or vertical)
        lcolor : text color
        lsize : text size
        olcolor : offset text color
        olsize : offset text size

    example:
        cbar(position=(0.1, 0.1), size=(0.7, 0.05), lsize=16, 
             lcolor='red', olsize=19, olcolor='b',direction='h') 
    '''
    pass


@redirect_to_aviewer
def savefig(filename):
    '''
    save figure as image

       savefig(filename)

       filename must be one of following
          .eps
          .pdf  (support multipage pdf)
          .svg
          .jpeg
          .png
          .gif  (animation gif)
    '''


@redirect_to_aviewer
def savedata(filename):
    '''
    save dataset as hdf file

      savedate(filename) # filename must be *.hdf
    '''

#
#   functions which are actually implemented here
#


@check_aviewer
def aviewer():
    return aviewer


@check_aviewer
def draw():
    '''
    draw draws the window contents. this command
    is intended to use with update('off') in script

    ex) ou = update()
        update('off')
        .... do some mupltiple plotting
        draw()
        update(ou)

    '''
    aviewer.draw()


@check_aviewer
def hold(val=None):
    '''
    hold controls if existing plots are deleted bofore
    adding a new one
       hold("on"), hold(1), hold(True)  -> hold is on
       hold("off"),hold(0), hold(False) -> hold is off

    '''
    if val is None:
        return globals()["_hold"]

    if isinstance(val, bool):
        globals()["_hold"] = val
    if isinstance(val, int):
        if (val == 1):
            globals()["_hold"] = True
        if (val == 0):
            globals()["_hold"] = False
    if isinstance(val, str):
        if (val.upper() == 'ON'):
            globals()["_hold"] = True
        if (val.upper() == 'OFF'):
            globals()["_hold"] = False


@check_aviewer
def update(val=None):
    '''
    update controls if it draws screen after interactive
    command.
    update('on'), update(1), update(True) : automatic update on
    update('off'), update(0), update(False) : automatic update off
    '''
    if val is None:
        return globals()["_update"]
    if isinstance(val, bool):
        globals()["_update"] = val
    if isinstance(val, int):
        if (val == 1):
            globals()["_update"] = True
            if not globals()["_update"]:
                draw()
        if (val == 0):
            globals()["_update"] = False
    if isinstance(val, str):
        if (val.upper() == 'ON'):
            if not globals()["_update"]:
                draw()
            globals()["_update"] = True

        if (val.upper() == 'OFF'):
            globals()["_update"] = False


@check_aviewer
def ipage():
    '''
    get current page number
    '''
    return globals()['aviewer'].ipage


@check_aviewer
def close(*args):
    '''
    close()
    clsee(1)   : close all figure window
    clsee(all) : close all figure window
    '''
    if len(args) == 0:
        m = getattr(aviewer, 'close')
        m()
    else:
        # close all viewer whith has close method
        # having close method indicates it inherit
        # BookViewerInteractive
        ifig_app = get_topwindow()
        for v in ifig_app.viewers[:]:
            if hasattr(v, 'close'):
                v.close()


def clear():
    ifig_app = get_topwindow()
    ifig_app.shell.clear()


def newbook(name='', basename=None):
    '''
    add a new book
    '''
    ifig_app = get_topwindow()
    book = ifig_app.proj.onAddBook(basename=basename)
    i_page = book.add_page()
    page = book.get_page(i_page)
    page.realize()
    page.add_axes()
    page.realize_children()
    page.set_area([[0, 0, 1, 1]])
    return book
#    ifigure.events.SendShowPageEvent(page)


def _open_book(book, viewer, **kwargs):
    ifig_app = get_topwindow()
    if ifig_app.find_bookviewer(book) is not None:
        ifig_app.find_bookviewer(book).Raise()
        ifig_app.aviewer = ifig_app.find_bookviewer(book)
        return
    ifigure.events.SendOpenBookEvent(book, w=ifig_app,
                                     viewer=viewer, useProcessEvent=True, **kwargs)
    ifigure.events.SendChangedEvent(book, w=ifig_app, useProcessEvent=True)
    ifigure.events.SendCanvasSelected(book.get_child(0), w=None,
                                      useProcessEvent=True)


def _get_book_by_number(parent, num, basename='book'):
    from ifigure.mto.fig_book import FigBook
    name = basename+str(num)
    if parent.has_child(name):
        book = parent.get_child(name=name)
        if not isinstance(book, FigBook):
            return None
        return book
    else:
        book = FigBook()
        ipage = book.add_page()
        book.get_page(ipage).add_axes()
        book.get_page(ipage).realize()
        book.get_page(ipage).set_area([[0, 0, 1, 1]])
        parent.add_child(name, book)
        ifigure.events.SendChangedEvent(book, w=wx.GetApp().TopWindow)
        return book


def figure(file='', book=None, viewer=None, **kwargs):
    '''
    create a new book and open it in a new figure window
         figure()  : open empty figure
         figure('***.bfz') : open book file
         figure(proj.book) : open FigBook object
         figure(1)  : open (or make) book1 under proj
         figure(1, parent)  : open (or make) book1 under parent
    '''
    from ifigure.widgets.book_viewer import BookViewer
    from ifigure.mto.fig_book import FigBook

    if isinstance(file, FigBook):
        book = file
        file = ''
    elif isinstance(file, int):
        num = file
        parent = book if book is not None else wx.GetApp().TopWindow.proj
        book = _get_book_by_number(parent, num)
        file = ''
        if book is None:
            return
    if book is None:
        book = newbook()
    if book.num_page() == 0:
        book.add_page()
    if viewer is None:
        viewer = BookViewer
    _open_book(book, viewer, **kwargs)
    viewer = wx.GetApp().TopWindow.find_bookviewer(book)
    if file == '':
        return viewer

    import os
    file = os.path.expanduser(file)
    if file.endswith('.bfz'):
        evt = None
        wx.CallAfter(viewer.onLoadBook, evt, file=file)
    return viewer

# def scope(type='direct'):


def scope(file='',  book=None,  viewer=None, **kwargs):
    '''
    open mdsscope
       scope() : open empty scope
       scope('***.pbz') : open book file
       scope('***.dat') : import *.dat as dwscope file
       scope(proj.book) : open FigBook object
    '''

    from ifigure.mto.fig_book import FigBook
    if isinstance(file, FigBook):
        book = file
        file = ''
    elif isinstance(file, int):
        num = file
        parent = book if book is not None else wx.GetApp().TopWindow.proj
        book = _get_book_by_number(parent, num, basename='scope')
        file = ''
        if book is None:
            return
    from ifigure.mdsplus.mdsscope import MDSScope
    if viewer is None:
        viewer = MDSScope
    if book is None:
        book = newbook(basename='scope')
    if book.num_page() == 0:
        book.add_page()
    book.get_page(0).set_nomargin(True)
    _open_book(book, viewer, **kwargs)
    viewer = wx.GetApp().TopWindow.find_bookviewer(book)
    if file == '':
        return viewer

    import os
    file = os.path.expanduser(file)
    if file.endswith('.dat'):
        wx.CallAfter(viewer.onImportDW, None, file=file)
    elif file.endswith('.bfz'):
        evt = None
        wx.CallAfter(viewer.onLoadBook, evt, file=file)
    return viewer


def videoviewer(file='', book=None):

    from ifigure.widgets.video_viewer import VideoViewer
    viewer = figure(file=file, book=book, viewer=VideoViewer)
    return viewer


def waveviewer(file='', book=None, nframe=30, sign=-1):

    from ifigure.widgets.wave_viewer import WaveViewer
    viewer = figure(file=file, book=book, viewer=WaveViewer)
    viewer.sign = sign
    viewer.nframe = nframe
    return viewer


def video(*args, **kargs):
    '''
    video viewer is to look video image (3D array)
    vidoe(x, y, z) or video(z)
    '''
    if len(args) == 1:
        z = args[0]
        x = np.arange(z.shape[-1])
        y = np.arange(z.shape[-2])
    elif len(args) == 3:
        z = args[0]
        x = args[1]
        y = args[2]
    else:
        raise ValueError

    v = videoviewer()
    o = v.image(*args, **kargs)
    v.goto_frame(0)

    return v


def futurize(obj=None, dryrun=False, verbose=False, unicode=True, stage1=True, stage2=True, help=False):
    '''
    futurize : an interface to PY2->PY3 conversion utility.

       it uses futurizer in future module. Default actin is to perform
       both stage1 and stage2 conversion.

       See more detail in
       https://python-future.org/futurize.html#forwards-conversion

       obj : either None, Folder, Script
           None: convert all scripts in project
           Folder: convert all scripts in folder
           Script: convert one script

       dryrun: does not save the conveted script.
       stage1: perform stage 1 conversion
       stage2: perform stage 2 conversion
    '''

    from ifigure.utils.future import futurizer as ft
    futurizer = ft()
    if help == True:
        futurizer.process_proj(dryrun=False, verbose=False, unicode=True, stage1=True, stage2=True,
                               help=True)
        return
    if obj is None:
        futurizer.process_proj(dryrun=dryrun, verbose=verbose, unicode=unicode,
                               stage1=stage1, stage2=stage2, help=False)
    from ifigure.mto.py_script import PyScript
    from ifigure.mto.py_code import PyFolder
    if isinstance(obj, PyScript):
        futurizer.process_script(obj, dryrun=dryrun, verbose=verbose, unicode=unicode,
                                 stage1=stage1, stage2=stage2, help=False)
    elif isinstance(obj, PyFolder):
        futurizer.process_folder(obj, dryrun=dryrun, verbose=verbose, unicode=unicode,
                                 stage1=stage1, stage2=stage2, help=False)
    else:
        pass


def scopenw(book):
    from ifigure.mdsplus.mdsscope_nw import MDSScopeNW
    return MDSScopeNW(book=book)


def tscope(file='',  book=None):
    ifig_app = get_topwindow()
    proj = ifig_app.proj

    if proj.setting.has_child('ts_worker'):
        workers = proj.setting.ts_worker
    else:
        file = os.path.join(ifigure.__path__[0], 'add_on',
                            'setting', 'module', 'mdsplus_worker.py')

        workers = proj.setting.add_absmodule(file)
        workers.rename('ts_worker')
        workers.setvar('translater', 'ts')
    v = scope(file=file, book=book, worker=workers)
    v.book.setvar('mdsplus_server', 'mdsplus.partenaires.cea.fr:8000:')


try:
    from petram.pi.shell_commands import petram
    has_petra = True
except:
    has_petra = False


def edit(file=''):
    app = wx.GetApp().TopWindow
    app.open_editor_panel()
    if file == '':
        app.script_editor.NewFile()
    else:
        import os
        file = os.path.expanduser(file)
        app.script_editor.OpenFile(file)
    if not app.isEditorAttached():
        app.script_editor.Raise()


def debug(command, *args):
    '''
     set and list debug level
     debug('set', level_name, level)
       or  debug('s', level_name, level)
     debug('list')
       or  debug('l', level_name, level)
    '''
    import ifigure.utils.debug
    if command.startswith('l'):
        for key in ifigure.utils.debug.debug_modes:
            print((key, ifigure.utils.debug.debug_modes[key]))
    elif command.startswith('s'):
        ifigure.utils.debug.set_level(args[0], args[1])


def profile(txt, *args):
    '''
    profile(txt)
    profile(txt, filename)

    run cProfile with locals in Shell
    '''
    ifig_app = get_topwindow()
    import cProfile
    l = ifig_app.shell.lvar
    cProfile.runctx(txt, {}, l, *args)


def profile_start():
    '''
    profiling start
    usage:
       pr = profile_start()
       ... do something
       profile_stop(pr)
    '''
    import cProfile
    print('starting profiler')
    pr = cProfile.Profile()
    pr.enable()
    return pr


def profile_stop(pr, sortby='cumulative'):
    '''
    profile_stop(pr, sortby='cumulative'):

    end profile
    sortby = 'cumulative', 'calls', 'cumtime',
             'file', 'filename', 'module',
             'ncalls', pcalls', 'line', 'name',
             'nfl', stdname', 'time', 'tottime'
    '''
    from six import StringIO
    import pstats
    pr.disable()
    # print 'stopped profiler'
    lsortby = ['cumulative', 'calls', 'cumtime',
               'file', 'filename', 'module',
               'ncalls', 'pcalls', 'line', 'name',
               'nfl', 'stdname', 'time', 'tottime']
    if not sortby in lsortby:
        print('invalid sortby')
        print(lsortby)
        return

    s = StringIO()
    sortby = sortby
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats()
    print((s.getvalue()))


def server(param=None, extra=None):
    '''
    server : control server mode
        server('on') : start server
        server('on', hostname) : start server
             in this case, server process is binded
             to hostname, allowing connection over
             network.
        server('off'): stop server
        server()     : show server information
    '''
    import ifigure.server
    server = ifigure.server.Server()
    if param == 'on':
        server.start(host=extra)
    elif param == 'off':
        server.stop()
    elif param is None:
        return server.info()


def importv(dest=None, path=''):
    '''
    import variables which was saved as pickled file

    '''
    import ifigure.utils.pickle_wrapper as pickle
    from ifigure.mto.py_code import PyData
    if dest is None:
        ifig_app = get_topwindow()
        dest = PyData()
        ifig_app.proj.add_child('data', dest)

    if path == '':
        open_dlg = wx.FileDialog(None, message="Select Data File",
                                 style=wx.FD_OPEN)
        if open_dlg.ShowModal() != wx.ID_OK:
            open_dlg.Destroy()
            return
        path = open_dlg.GetPath()
        open_dlg.Destroy()
        if path == '':
            return
    fid = open(path, 'r')
    data = pickle.load(fid)
    fid.close()

    for key in data:
        dest.setvar(key, data[key])

    ifigure.events.SendChangedEvent(dest, w=ifig_app, useProcessEvent=True)
    return dest


def exportv(variables, names, path=''):
    '''
    export variables as pickled file

    example: export([x, y, z], ['x', 'y', 'z'])
    note: one can make data_tree object and save it as project
          or export subtree.

    '''
    import cPickle as pickle
    save_dlg = wx.FileDialog(None, message="Enter Data File Name",
                             defaultDir=os.getcwd(),
                             style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
    if save_dlg.ShowModal() != wx.ID_OK:
        save_dlg.Destroy()
        return
    path = save_dlg.GetPath()
    save_dlg.Destroy()
    if path == '':
        return

    fid = open(path, 'w')
    d = {n: v for v, n in zip(variables, names)}
    pickle.dump(d, fid)
    fid.close()


def quit():
    '''
    quit piScope
    '''
    ifig_app = get_topwindow()
    ifig_app.onQuit()


def glinfo():
    '''
    show OpenGL information
    '''
    try:
        import OpenGL
        import OpenGL.GL
    except ImportError:
        print("OpenGL not avaiable")
        return
    print('OpenGL Ver. : ' + OpenGL.GL.glGetString(OpenGL.GL.GL_VERSION).decode())
    print('GS Lang Ver.: ' +
          OpenGL.GL.glGetString(OpenGL.GL.GL_SHADING_LANGUAGE_VERSION).decode())
    print('Vendor      : ' + OpenGL.GL.glGetString(OpenGL.GL.GL_VENDOR).decode())
    print('Renderer    : ' + OpenGL.GL.glGetString(OpenGL.GL.GL_RENDERER).decode())


def setupmodel(package='', root='', path='setup_scripts', model=None,
               del_scripts=True):
    '''
    Utility command to setup simulation model. It uses
    mercurial repositories to store skelton scripts (and
    other pieces).

    setupmodel()
    setupmodel(package = '', root = '', path = 'setup_scripts', model=None,
               del_scripts = True, ):
    input:
         package: package name
         root   : root hg repository
         path   : paht to setup_scripts in repo
         model  : destination
    return :
         model object

    example:
         setupmodel()  : open dialog to pick model package
         setupmodel('genray_cql3d', '~/hg_root/ppkgs')
                       : setup genray_cql3d model pakcage using data
                         in hg repo at ~/hg_root/ppkags
    '''
    from ifigure.utils.model_setup_tools import setup
    return setup(package=package, model=model, root=root, path=path,
                 del_scripts=del_scripts)


def autoplay(viewer=None, interval=0.2):
    #timer = wx.Timer()
    # self.viewer.Bind(

    if viewer is None:
        viewer = aviewer
    if viewer is None:
        return

    def step_viewer(player=viewer, interval=interval):
        ipage = viewer.ipage
        num_page = viewer.book.num_page()
        if ipage == num_page-1:
            ipage = 0
        else:
            ipage = ipage + 1
        viewer.show_page(ipage)
        if viewer.timer is not None:
            viewer.timer.Start(interval*1000., oneShot=True)

    viewer.timer = wx.Timer(viewer)
    viewer.Bind(wx.EVT_TIMER, step_viewer)
    if viewer.isPropShown():
        viewer.toggle_property()

    viewer.timer.Start(interval*1000., oneShot=True)

    ifig_app = get_topwindow()
    x = ifig_app.shell.raw_input('stop?')
    viewer.timer.Stop()
    viewer.timer = None


# get_shellvar/put_shellvar is to manipulate shell variable
# from client
def get_shellvar(name):
    var = wx.GetApp().TopWindow.shell.lvar
    if name in var:
        return var[name]


def put_shellvar(name, value):
    var = wx.GetApp().TopWindow.shell.lvar
    var[name] = value

#
#  TODO (following functions needs to be revised
#


def get_page(ipage=None):
    return aviewer.get_page(ipage=ipage)


def get_axes(ipage=None, iaxes=None):
    return aviewer.get_axes(ipage=ipage, iaxes=iaxes)


def twinc():
    if aviewer is None:
        return
    fig_p = get_page(ipage=None)
    if fig_p is None:
        print("no page exists. use addpage() to create a page")
        return
    axes = get_axes(ipage=None, iaxes=None)
    axes.add_axis_param(dir='c')
    axes.set_bmp_update(False)
    draw()
    ifigure.events.SendChangedEvent(axes, w=aviewer, useProcessEvent=True)



import wx
import wx.stc as stc


class DlgMessageScroll(wx.Dialog):
    def __init__(self, *args, **kargs):
        wx.Dialog.__init__(self, *args,
                           style=wx.CAPTION |
                           wx.CLOSE_BOX |
                           wx.MINIMIZE_BOX |
                           wx.RESIZE_BORDER | wx.STAY_ON_TOP)

        vbox = wx.BoxSizer(wx.VERTICAL)

        panel = wx.Panel(self, wx.ID_ANY)
        panel.SetSizer(vbox)
       # self.tip = wx.StaticText(panel, -1,
       #                          style=wx.TE_MULTILINE)#|wx.TE_READONLY|wx.VSCROLL)
       # self.tip.SetBackgroundColour((255,255,0))
        self.tip = stc.StyledTextCtrl(panel, -1)
        self.tip.SetMarginWidth(0, 0)
        self.tip.SetMarginWidth(1, 0)
        self.tip.SetMarginWidth(2, 0)
        self.tip.SetReadOnly(True)
        self.tip.StyleSetBackground(wx.stc.STC_STYLE_DEFAULT, (190, 190, 190))

        vbox.Add(self.tip, 1, wx.EXPAND | wx.ALL, 5)
        button1 = wx.Button(panel, label="O.K.")
        vbox.Add(button1, 0, wx.ALIGN_RIGHT | wx.RIGHT, 20)
        self.Bind(wx.EVT_CLOSE, self.onWindowClose)
        self.Bind(wx.EVT_BUTTON, self.onOK, button1)
        #panel.Bind(wx.EVT_SET_FOCUS, self.onSetFocus)
        #panel.Bind(wx.EVT_KILL_FOCUS, self.onKillFocus)
        #self.tip.Bind(wx.EVT_SET_FOCUS, self.onSetFocus)
        #self.tip.Bind(wx.EVT_KILL_FOCUS, self.onKillFocus)
        # self.Hide()

#    def update(self, func):
    def update(self, error):
        #        if hasattr(func, '__doc__'):
        try:
            #                self.tip.SetLabel(func.__doc__)
            self.tip.SetReadOnly(False)
            self.tip.SetText(error)
            self.tip.StyleClearAll()
            self.tip.SetReadOnly(True)
        except:
            pass

    def onOK(self, e):
        self.Close()

    def onWindowClose(self, e):
        e.Skip()



from __future__ import print_function
import __builtin__
openfiles = set()
oldfile = __builtin__.file


class newfile(oldfile):
    def __init__(self, *args):
        self.x = args[0]
        print("### OPENING %s ###" % str(self.x))
        oldfile.__init__(self, *args)
        openfiles.add(self)

    def close(self):
        print("### CLOSING %s ###" % str(self.x))
        oldfile.close(self)
        openfiles.remove(self)


oldopen = __builtin__.open


def newopen(*args):
    return newfile(*args)


__builtin__.file = newfile
__builtin__.open = newopen


def printOpenFiles():
    print("### %d OPEN FILES: [%s]" % (
        len(openfiles), ", ".join(f.x for f in openfiles)))



'''
    Video Viewer

    VideoViewer is a type of dynamic data generation, in which
    3D arrays is treated as video data.

     See image_video_demo() in ifigure.plot_demos for example

    VideoViewerMode is a mix-in class which repurpose multipage 
    feature of BookViewer to realize daynamic data generatation and 
    drawing. This class is also used in wave_viewer, in which
    frame data is generater dynamically using phasor.

'''
from ifigure.mto.fig_image import FigImage
import wx
from ifigure.widgets.book_viewer import BookViewer, BookViewerFrame, ID_KEEPDATA
import ifigure.events
import weakref
import numpy as np
from .videoplayer_buttons import VideoplayerButtons, add_player_btn


class VideoViewerMode(object):
    ID_PLAYERBUTTON = wx.NewIdRef(count=1)

    def __init__(self, *args, **kwargs):
        self._video_obj = weakref.WeakSet()
        self._video_page = -1
        if not hasattr(self, '_playerbtn'):
            self._playerbtn = None
        self._playinterval = 100
        self._playloop = False
        self._is_playing = 0  # 0: stop, 1: forward, -1: backward

    def add_all_video_obj(self):
        raise NotImplementedError(
            'VideoViewerMode::add_all_video_obj should be overwritten')

    def reset_video_obj_set(self):
        self._video_obj = weakref.WeakSet()

    def add_video_obj(self, figobj):
        self._video_obj.add(figobj)

    def set_window_title(self):
        if self.book is None:
            self.SetTitle('')
            return
        title = self.book.get_full_path()+'(frame '+str(self._video_page+1)+')'
        self.SetTitle(title)

    def onGoToFrame(self, evt):
        from ifigure.utils.edit_list import DialogEditList
        list6 = [("Frame number (1-" + str(self.num_page()) + ")",
                  str(self._video_page), 0, None), ]
        value = DialogEditList(list6, modal=True,
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
                               tip=None,
                               parent=self,)

        if value[0] is True:
            ipage = int(value[1][0])
        else:
            return
        self.goto_frame(ipage)
        self.property_editor.onTD_ShowPage(evt)

    def goto_frame(self, ipage):
        self.UpdateImage(ipage)
        self.canvas.exit_layout_mode()
        self.set_window_title()
        wx.CallAfter(self.draw)

    def _onTD_ShowPage(self, evt):
        from ifigure.mto.fig_page import FigPage
        dt = evt.GetTreeDict()  # dt fig_page
        if isinstance(dt, FigPage):
            self.book.set_open(False)
            dt.get_root_parent().set_pbook(dt.get_parent())
            self.book = dt.get_parent()
            self.ipage = 0
        elif dt == self.book:
            ipage = self._video_page
            pass
        else:
            return
        ipage = self._video_page
        num_page = self.num_page()
        if evt.inc == '-1':
            if self._playloop:
                if self._video_page == 0:
                    ipage = num_page - 1
                else:
                    ipage = self._video_page - 1
            else:
                ipage = max([self._video_page-1, 0])
        elif evt.inc == '1':
            if self._playloop:
                if self._video_page == num_page-1:
                    ipage = 0
                else:
                    ipage = self._video_page+1
            else:
                ipage = min([self._video_page+1, num_page-1])
        elif evt.inc == 'first':
            ipage = 0
        elif evt.inc == 'last':
            ipage = num_page-1
        else:
            pass
        self.goto_frame(ipage)
        self.property_editor.onTD_ShowPage(evt)

    def onTD_ShowPage(self, evt):
        if not evt.BookViewerFrame_Processed:
            self._onTD_ShowPage(evt)
            evt.BookViewerFrame_Processed = True
            evt.SetEventObject(self)
            if self._is_playing != 0:
                wx.CallLater(self._playinterval, self.onTime)
        evt.Skip()

    def onNextPage(self, evt):
        self.onNextVideoPage(evt)

    def onPrevPage(self, evt):
        self.onPrevVideoPage(evt)

#    def onActivate(self, evt):
#        if self._playerbtn is not None:
#            self._playerbtn.Raise()
    def save_animgif(self, filename='animation.gif',
                     show_page=None, duration=None, dither=None, pages='all'):
        def show_page(args):
            k = args[0]
            self.goto_frame(k)
            self.draw()
        BookViewer.save_animgif(self, filename=filename,
                                show_page=show_page, duration=duration, dither=dither, pages=pages)

    def save_animpng(self, filename='animation.png',
                     show_page=None, duration=None, pages='all'):
        def show_page(args):
            k = args[0]
            self.goto_frame(k)
            self.draw()
        BookViewer.save_animpng(self, filename=filename,
                                show_page=show_page, duration=duration, pages=pages)

    def save_multipdf(self, filename='figure_allpage.pdf',
                      show_page=None):
        def show_page(*args):
            import time
            k = args[0]
            self.goto_frame(k)
            time.sleep(0.01)
            self.draw()

        BookViewer.save_multipdf(self, filename=filename,
                                 show_page=show_page)

    def add_bookmenus(self, editmenu, viewmenu):

        self.add_menu(editmenu, VideoViewerMode.ID_PLAYERBUTTON,
                      "Go to Frame...",
                      "select frame number",
                      self.onGoToFrame)
        editmenu.AppendSeparator()
        self.add_menu(editmenu, ID_KEEPDATA,
                      "Keep Book in Tree",
                      "Book data is kept in tree when this window is closed",
                      self.onKeepData,
                      kind=wx.ITEM_CHECK)

        self._mm_player = self.insert_menu(viewmenu, 1, wx.ID_ANY,
                                           "Player Panel", "toggle player button panel",
                                           self.onTogglePlayerButton,
                                           kind=wx.ITEM_CHECK)
        self._playerbtn = add_player_btn(self)
        # wx.GetApp().add_palette(self._playerbtn)
        self._mm_player.Check(True)
        self.add_menu(viewmenu, BookViewerFrame.ID_PM[4],
                      "Next Page",  "next page",
                      self.onNextVideoPage)
        self.add_menu(viewmenu, BookViewerFrame.ID_PM[5],
                      "Previous Page", "previous page",
                      self.onPrevVideoPage)

    def onTogglePlayerButton(self, evt):
        if self._mm_player.IsChecked():
            if self._playerbtn is None:
                self._playerbtn = add_player_btn(self)
                # wx.GetApp().add_palette(self._playerbtn)
        else:
            # wx.GetApp().rm_palette(self._playerbtn)
            self._playerbtn.Destroy()
            self._playerbtn = None

    def onPlayerButtonClose(self):
        self._mm_player.Check(False)
        self._playerbtn = None

    def goto_last(self):
        ifigure.events.SendShowPageEvent(self.book, self, 'last')

    def goto_first(self):
        ifigure.events.SendShowPageEvent(self.book, self, 'first')

    def step_fwd(self):
        ifigure.events.SendShowPageEvent(self.book, self, '1')

    def step_rev(self):
        ifigure.events.SendShowPageEvent(self.book, self, '-1')

    def onTime(self):
        num_page = self.num_page()
        if self._is_playing != 0:
            if (self._video_page == 0 and self._is_playing == -1 and
                    not self._playloop):
                self._is_playing = 0
                if self._playerbtn is not None:
                    self._playerbtn.reset_btn_toggle_bitmap()
                return
            if (self._video_page == num_page-1 and self._is_playing == 1 and
                    not self._playloop):
                if self._playerbtn is not None:
                    self._playerbtn.reset_btn_toggle_bitmap()
                self._is_playing = 0
                return
            ifigure.events.SendShowPageEvent(
                self.book, self, str(self._is_playing))

    def play_fwd(self):
        self._is_playing = 1
        wx.CallLater(self._playinterval, self.onTime)

    def play_rev(self):
        self._is_playing = -1
        wx.CallLater(self._playinterval, self.onTime)

    def stop_play(self):
        self._is_playing = 0

    def videoviewer_config(self):
        from ifigure.utils.edit_list import DialogEditList
        l = [
            ["Interval(sec.)", str(float(self._playinterval)/1000.),
             0, {'noexpand': True}],
            [None, self._playloop, 3, {"text": "Loop", "noindent": None}],
        ]
        value = DialogEditList(l, parent=self,
                               title='Player Config.',
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        if not value[0]:
            return
        interval = min([50, int(float(value[1][0])*1000.)])
        self._playinterval = int(interval)
        self._playloop = bool(value[1][1])


class FigImageVideo(FigImage):
    def set_video_data(self, value):
        self._video = value

    def get_nframe(self):
        return self._video.shape[0]

    def save_data2(self, data=None):
        data['FigImageVideo'] = (1, {'_vidoe': self._video})
        data = super(FigImageVideo, self).save_data2(data)
        return

    def load_data2(self, data=None):
        d = data['FigImageVideo']
        super(FigImageVideo, self).load_data2(data)
        self._video = d[1]['_video']

    def show_videoframe(self, i):
        self.setp('z', self._video[i])
        x, y, z = self.getp(("x", "y", "z"))
        xp, yp, zp = self.interp_image(x, y, z)
        a = self._artists[0]
        a.set_array(zp)
        setattr(a.get_array(), '_xyp', (xp, yp))
        self.set_bmp_update(False)


def convert_figobj(obj):
    if obj.__class__ == FigImage:
        obj.__class__ = FigImageVideo


class VideoBookPlayer(VideoViewerMode, BookViewer):
    def __init__(self, *args, **kwargs):
        VideoViewerMode.__init__(self, *args, **kwargs)
        BookViewer.__init__(self, *args, **kwargs)

    def onResize(self, evt):
        BookViewer.onResize(self, evt)
        if self._playerbtn is not None:
            self._playerbtn.place_right_bottom()

    def call_draw_after_resize(self):
        if self._playerbtn is not None:
            self._playerbtn.place_right_bottom()
        BookViewer.call_draw_after_resize(self)

    def onPaste(self, e):
        BookViewer.onPaste(self, e)
        self.add_all_video_obj()

    def add_all_video_obj(self):
        raise NotImplementedError('Need to implement in subclass')


class VideoViewer(VideoBookPlayer):
    def __init__(self, *args, **kwargs):
        self._video_obj = []
        kwargs['isattachable'] = False
        super(VideoViewer, self).__init__(*args, **kwargs)

        if self.book is not None:
            self.add_all_video_obj()

    def image(self, *args, **kwargs):
        '''
        image(z)
        or
        image(x, y, z)
        '''
        if len(args) == 1:
            z = args[0]
            x = np.arange(z.shape[-1])
            y = np.arange(z.shape[-2])
        elif len(args) == 3:
            x = args[0]
            y = args[1]
            z = args[2]
        if z.ndim != 3:
            raise ValueError('z data should have ndim = 3')

        o = BookViewer.image(self, x, y, z[0], **kwargs)
        if o is None:
            return
        convert_figobj(o)
        o.set_video_data(z)
        self.add_video_obj(o)
        return o

    def add_all_video_obj(self):
        self.reset_video_obj_set()
        for obj in self.book.walk_tree():
            if isinstance(obj, FigImageVideo):
                self.add_video_obj(obj)

    def num_page(self):
        return min([o.get_nframe() for o in self._video_obj])

    def onNextVideoPage(self, e):
        num_page = self.num_page()
        if self._video_page == num_page-1:
            return
        ifigure.events.SendShowPageEvent(self.book, self, '1')

    def onPrevVideoPage(self, e):
        if self._video_page == 0:
            return
        ifigure.events.SendShowPageEvent(self.book, self, '-1')

#    def onTD_ShowPage(self, evt):
#        if not evt.BookViewerFrame_Processed:
#           VideoViewerMode.onTD_ShowPage(self, evt)
#           evt.BookViewerFrame_Processed = True
#           evt.SetEventObject(self)
#        evt.Skip()

    def onAttachFigure(self, e):
        pass

    def adjust_attach_menu(self):
        pass

    def UpdateImage(self, i):
        for obj in self._video_obj:
            obj.show_videoframe(i)
        self._video_page = i



from __future__ import print_function
import numpy as np
import collections


def rec2dict(arr, cls=collections.OrderedDict):
    if not hasattr(arr, 'flatten'):
        return arr
    if not hasattr(arr.flatten()[0], 'dtype'):
        return arr
    elif arr.dtype.names is not None:
        r = cls()
        for name in arr.dtype.names:
            print(name)
            print(isinstance(arr[name], np.recarray))
            if isinstance(arr[name], np.recarray):
                r[name] = rec2dict(arr[name])
            elif (isinstance(arr[name], np.ndarray) and
                  arr[name].ndim == 1
                  and arr[name].shape[0] == 1 and
                  isinstance(arr[name][0], np.recarray)):
                r[name] = rec2dict(arr[name][0], cls=cls)
            else:
                r._var0[name] = arr[name]
        return r
    elif arr.dtype.hasobject:
        if arr.ndim == 1 and arr.shape[0] == 1:
            return rec2dict(arr[0])
        else:
            arr2 = arr.flatten()
            v = np.array([None]*len(arr2))
            for k, x in enumerate(arr2):
                v[k] = rec2dict(arr2[k])
            return v.reshape(arr.shape)
    else:
        return arr



from ifigure.mto.fig_obj import FigObj, mask_negative
from ifigure.mto.fig_image import FigImage
from ifigure.mto.triangle_plots import TrianglePlots
from ifigure.utils.cbook import ProcessKeywords
from matplotlib.tri import Triangulation
import numpy as np
import ifigure.events


class FigTripcolor(FigImage, TrianglePlots):
    '''
    tripcolor(x, y, z)
    tripcolor(tri, x, y, z)
    '''
    default_rasterized = True

    def __new__(self, *args, **kywds):
        if len(args) == 4:
            tri = args[0]
            args = args[1:]
        else:
            tri = None
        kywds['use_tri'] = True
        shading = kywds.pop('shading', "flat")
        mask = kywds.pop('mask', None)
        obj = FigImage.__new__(self, *args, **kywds)
        obj.setp('shading', shading)
        obj.setp('mask', mask)
        obj.setvar('shading', shading)
        obj.setvar('mask', mask)
        obj.setvar("tri", tri)
        return obj

    def __init__(self, *args,  **kywds):
        if len(args) == 4:
            tri = args[0]
            args = args[1:]
        else:
            tri = None
        FigImage.__init__(self, *args, **kywds)
        TrianglePlots.__init__(self)
        self._tri = tri

    @classmethod
    def isTripColor(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'tripcolor'

    @classmethod
    def attr_in_file(self):
        return (["alpha", "shading", "mask"] +
                super(FigImage, self).attr_in_file())

    @classmethod
    def property_in_palette(self):
        return ["cmap", "tripcolor_shading", "alpha_2"]
#    def set_shading(self, value, a):
#        self.setp('shading', value)
#        a.set_shading(value)
#    def get_shading(self, a):
#        return self.getp('interp')

    def onResize(self, evt):
        self.set_bmp_update(False)

        # self.del_artist(delall=True)
        # self.delp('loaded_property')
        #self.setp('use_var', False)
        # if hasattr(self, "_data_extent"):
        #   self._data_extent = None
        # self.generate_artist()

    def handle_axes_change(self, evt=None):
        return super(FigImage, self).handle_axes_change(evt)

    def set_alpha(self, value, a):
        a.set_alpha(value)
        cax = self.get_caxisparam()
        cax.set_crangeparam_to_artist(a)
        self.setp('alpha', value)
        self.set_bmp_update(False)

    def get_alpha(self, a):
        return a.get_alpha()

    def get_crange(self, crange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        if len(self._artists) == 0:
            return FigImage.get_crange(self, crange=crange,
                                       xrange=xrange, yrange=yrange,
                                       scale=scale)
        else:
            zt = self._artists[0].get_array()
            if scale == 'log':
                zt = mask_negative(zt)
            crange = self._update_range(crange,
                                        (min(zt), max(zt)))
        return crange

    def _eval_xyz(self):
        if self._tri is None:
            self._tri = self.getvar('tri')
        return FigImage._eval_xyz(self)

    def picker_a(self, artist, evt):
        '''
        this picker is faster since it judges based on
        x, y nodes. It does not look if the mouse is 
        inside the path. A user has to click the corner
        of triangles...
        '''
        x, y = self.getp(('x', 'y'))
        ptx = np.vstack((x.flatten(), y.flatten()))
        t = self._artists[0].axes.transData
        ptx = t.transform(ptx.transpose())
        dist = np.sqrt((ptx[:, 0] - evt.x) ** 2 + (ptx[:, 1] - evt.y)**2)
        if np.min(dist) < 5:
            self._pick_pos = [evt.xdata, evt.ydata]
            return True,  {'child_artist': artist}
        else:
            return False, {}



import six

if six.PY2:
    import cPickle as pickle
else:
    import pickle

from ifigure.ifigure_config import pickle_protocol

def dump(data, fid, **kwargs):
    if six.PY2:
        pickle.dump(data, fid)
    else:
        kwargs['protocol'] = pickle_protocol
        if pickle_protocol == 2:
            kwargs['fix_imports'] = True
        pickle.dump(data, fid, **kwargs)

def load(fid):
    if six.PY2:
        return pickle.load(fid)
    else:
        return pickle.load(fid, fix_imports=True, encoding='latin1')

dumps = pickle.dumps
loads = pickle.loads



#!/usr/bin/python

from __future__ import print_function
import socket
import threading
from six.moves import socketserver
import paramiko
import os
import MDSplus
import binascii
import sys
import cPickle as pickle


class ThreadedTCPRequestHandler(socketserver.BaseRequestHandler):

    #

    def handle(self):
        data = self.request.recv(1024)
        #cur_thread = threading.current_thread()
        com = data[0]
        param = data[1:]

        response = ''
        print(("Request ", data))
        if com == 'c':  # connection request
            tree, shot, node = param.split(',')
            try:
                self.t = MDSplus.Tree(tree, int(shot))
                tn = self.t.getNode(node)
                self.t.setDefault(tn)
                response = 'ok'
            except Exception:
                response = str(sys.exc_info()[1])+','+str(sys.exc_info()[2])

        if com == 'f':  # connection request
            a = param.split(',')
            tree = a[0]
            shot = a[1]
            node = a[2]
            expr = ','.join(a[3:])
            try:
                self.t = MDSplus.Tree(tree, int(shot))
                if node.strip() != '':
                    tn = self.t.getNode(node)
                    self.t.setDefault(tn)
                r = MDSplus.Data.compile(expr).evaluate().data()
                sr = pickle.dumps(r)
                response = binascii.b2a_hex(sr)
                print(("sending data (length)", len(
                    response), 'original :', type(sr)))
                print(type(binascii.a2b_hex(response)))

            except Exception:
                response = str(sys.exc_info()[1])+','+str(sys.exc_info()[2])

        if com == 'v':  # mdsvalue
            r = MDSplus.Data.compile(param).evaluate().data()
            sr = pickle.dumps(r)
            response = binascii.b2a_hex(sr)
            print(("sending data (length)", len(
                response), 'original :', type(sr)))
            print(type(binascii.a2b_hex(response)))
        #response = "{}: {}".format(cur_thread.name, data)

        self.request.sendall(response+'\n')


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass


if __name__ == "__main__":
    # Port 0 means to select an arbitrary unused port
    HOST, PORT = os.getenv("HOSTNAME"), 10002

    server = ThreadedTCPServer((HOST, PORT), ThreadedTCPRequestHandler)
    ip, port = server.server_address

    print((ip, port))
    # Start a thread with the server -- that thread will then start one
    # more thread for each request
    server_thread = threading.Thread(target=server.serve_forever)
    # Exit the server thread when the main thread terminates
    server_thread.daemon = True
    server_thread.start()

    # usr=os.getenv("USER")
    from ifigure.utils.get_username import get_username
    usr = get_username()
    print(("Server loop running in thread:", server_thread.name))

    raw_input()

    server.shutdown()



from __future__ import print_function

from ifigure.utils.cbook import Write2Main
from ifigure.mto.py_code import PyData
from ifigure.mto.py_code import PyCode
import wx
import os
import re
import string
import sys
from numpy import *
from collections import OrderedDict

######################################################
#         Setting for module file for py_module
#
#   General rule:
#      This file will be automatically loaded when
#      py_module object is created. Also, py_module
#      keeps track the file modification time. If
#      the file is updated, it will be automaticaaly
#      reloaded.


#      Strong recommendation : make module "independent".

#      Py_Modules does not check the dependency of
#      modules.
#      If moduels used in Py_Modules depends on
#      each other by for example module variable,
#      it will cause complicate  module loading
#      order-dependency problem.
#
#   name of module
module_name = 'dwscope_subs'
#   module_evt_handler
#   functions which can be called from project tree
#
#    (menu name, function name, a flat to call skip()
#     after running function)
#
#   By default these function should return None
#   or True
#   if it return False, ifigure stops exectuion at
#   this module
#
module_evt_handler = [("Load File", "onLoadFile", True)]
#
#   A function called when py_module is initialized
#
module_init = None
######################################################


def import_dwscope(file):

    f = open(file, 'r')

    d = {}
    while 1:
        line = f.readline()
        if not line:
            break
        line = line.rstrip("\r\n")

        pos = line.find(':')
        command = line[:pos]
        data = line[pos+1:]
        print(line)
        arr = command.split('.')
        if len(arr) == 3:
            if arr[1] not in d:
                d[arr[1]] = {}
            d[arr[1]][arr[2]] = data
        if len(arr) == 2:
            if arr[1] not in d:
                d[arr[1]] = {}
            d[arr[1]] = data

    f.close()

    num_c = int(d["columns"])
    num_r = [0]*num_c
    return d
    for i in range(num_c)+1:
        key = 'rows_in_column_'+str(i)
        num_r[i-1] = int(d[key])

    d2 = {"num_c": num_c,
          "num_r": num_r}

    for key in d:
        if key[:4] == 'plot':
            d2[key[5:]] = d[key]
        if key[:6] == 'global':
            d2["global"] = d[key]

    d2["title"] = d["title"]
    d2["title_event"] = d["title_event"]
    return d2


def onLoadFile(self_obj):
    open_dlg = wx.FileDialog(None, message="Select DWscope",
                             wildcard='*.dat', style=wx.FD_OPEN)
    if open_dlg.ShowModal() != wx.ID_OK:
        open_dlg.Destroy()
        return False

    file = open_dlg.GetPath()
    open_dlg.Destroy()
    for name, child in self_obj.get_children():
        child.destroy()
    self_obj.setvar("filename", file)

    value = import_dwscope(file)
    self_obj.setvar("value", value)



import wx
import os
import weakref
import ifigure
from ifigure.utils.edit_list import DialogEditListTab
import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('dlg_preference')


class PrefComponent(object):
    def __init__(self, cname='Sample'):
        self.name = cname

    def get_dialoglist(self):
        list1 = [["Sample", 'preferred setting', 0, None], ]
        hint1 = [None]
        return list1, hint1

    def set_dialog_result(self, data):
        dprint1('set_dialog_result should be overwritten')

    def save_setting(self):
        dprint1('save_setting should be overwritten')


def dlg_preference(components, parent):

    lists = []
    hints = []
    tabs = []
    for c in components:
        l, h = c.get_dialoglist()
        lists.append(l)
        hints.append(h)
        tabs.append(c.name)

    value = DialogEditListTab(tabs, lists, tip=hints, parent=parent,
                              style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

    if value[0]:
        for i, v in enumerate(value[1]):
            components[i].set_dialog_result(v)



from __future__ import print_function
from ifigure.mto.fig_obj import FigObj, mask_negative
from ifigure.widgets.canvas.file_structure import *
from ifigure.mto.axis_user import XUser, YUser
from ifigure.mto.fig_obj import set_mpl_all, get_mpl_first
import ifigure
import os
import ifigure.utils.cbook as cbook
from ifigure.utils.geom import transform_point
import ifigure.widgets.canvas.custom_picker as cpicker
from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.widgets.undo_redo_history import UndoRedoArtistProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod
import numpy as np
import weakref

from ifigure.utils.args_parser import ArgsParser
import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('FigAxspan')


def _special_xy_check(v):
    # special check for x and y
    try:
        if (len(v) == 2):
            x0 = v[0]
            x1 = v[1]
            if ((isinstance(x0, float) or isinstance(x0, int)) and
                    (isinstance(x1, float) or isinstance(x1, int))):
                return [[x0, x1]], True
        for item in v:
            if len(item) != 2:
                flag = False
    except:
        flag = False
        return [], flag
    return v, True


def _make_xy_x(x):
    return np.array([[x[0], 0], [x[0], 1], [x[1], 1], [x[1], 0], [x[0], 0]])


def _make_xy_y(y):
    return np.array([[0, y[0]], [1, y[0]], [1, y[1]], [0, y[1]], [0, y[0]]])


def _sort_data(data):
    return [[min(item), max(item)] for item in data]


def _is_data_clean(data):
    for item in data:
        if not (np.isfinite(item[0]) and np.isfinite(item[1])):
            return False
    return True


def _clean_data(data):
    return [item for item in data
            if np.isfinite(item[0]) and np.isfinite(item[1])]


class FigAxspan(FigObj, XUser, YUser):
    def __new__(cls, *args, **kywds):
        def set_hidden_vars(obj):
            obj._objs = []
            obj._drag_backup = None
            obj._drag_delta = None
            obj._drag_mode = 0
            obj._data_extent = None
            return obj
        if 'src' in kywds:
            obj = FigObj.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            return obj

        p = ArgsParser()
        p.add_opt('x', [], ['can_ndreal_array|nonstr', 'dynamic', 'empty'])
        p.add_opt('y', [], ['can_ndreal_array|nonstr', 'dynamic', 'empty'])
        p.add_key('draggable', True)
        p.set_ndconvert("x", "y")
        v, kywds, d, flag = p.process(*args, **kywds)

        # special check for x and y
        if not 'x' in v:
            raise ValueError('can not parse arguments')
        if not 'y' in v:
            raise ValueError('can not parse arguments')
        v["x"], flag2 = _special_xy_check(v["x"])
        v["y"], flag3 = _special_xy_check(v["y"])

        if not flag or not flag2 or not flag3:
            raise ValueError('Failed when processing argument')

        obj = FigObj.__new__(cls, *args, **kywds)

        obj = set_hidden_vars(obj)
        for name in v:
            obj.setvar(name, v[name])
        obj.setvar("kywds", kywds)

        return obj

    def __init__(self, *args,  **kywds):
        XUser.__init__(self)
        YUser.__init__(self)

        args = []
        if 'src' not in kywds:
            kywds = self.getvar("kywds")
        FigObj.__init__(self, *args, **kywds)

    @classmethod
    def isFigAxspan(self):
        return True

    @classmethod
    def can_have_child(self, child=None):
        return False

    @classmethod
    def get_namebase(self):
        return 'axspan'

    @classmethod
    def property_in_file(self):
        return ["facecolor", "edgecolor", "linestyle",
                "linewidth", "alpha"]

    @classmethod
    def property_in_palette(self):
        return ["fill_2", "facecolor_2",
                "edgecolor_2", "plinestyle_2",
                "linewidth_2", "alpha_2"]

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'plot1.png')
        return [idx1]

    def isDraggable(self):
        return self._var["draggable"]

    def set_parent(self, parent):
        XUser.unset_ax(self)
        YUser.unset_ay(self)
        super(FigAxspan, self).set_parent(parent)
        XUser.get_xaxisparam(self)
        YUser.get_yaxisparam(self)

    def args2var(self):
        names0 = self.attr_in_file()
        names = ["x", "y"]
        use_np = [False]*2
        names = names + names0
        use_np = use_np + [False]*len(names0)
        values = self.put_args2var(names, use_np)

        x = values[0]
        y = values[1]
        if x is None:
            return False
        if y is None:
            return False
        x, flag = _special_xy_check(x)
        if not flag:
            return False
        y, flag = _special_xy_check(y)
        if not flag:
            return False
        try:
            if not _is_data_clean(x):
                return False
            if not _is_data_clean(y):
                return False
        except:
            return False
        x = np.array(_sort_data(x))
        y = np.array(_sort_data(y))
        self.setp("x", x)
        self.setp("y", y)
        return True

    def generate_artist(self):
        # this method generate artist
        # if artist does exist, update artist
        # based on the information specifically
        # managed by fig_obj tree. Any property
        # internally managed by matplotlib
        # does not change
        #           container=self._parent._artists[0]
        container = self.get_container()
        if self.isempty() is False:
            return
#              self.del_artist(delall=True)

        x, y = self._eval_xy()
        for x1 in x:
            self._artists.append(container.axvspan(x1[0], x1[1],
                                                   **(self._var["kywds"])))
        for y1 in y:
            self._artists.append(container.axhspan(y1[0], y1[1],
                                                   **(self._var["kywds"])))

        self.set_rasterized()
        lp = self.getp("loaded_property")
        if lp is not None and len(lp) > 0:
            for i, a in enumerate(self._artists):
                #              for i in range(0, len(lp)):
                #                  if len(self._artists) > i:
                lpp = lp[i] if len(lp) > i else lp[-1]
                alpha = lpp['alpha']
#                   del lp[i]['alpha']
                self.set_artist_property(a,
                                         {"alpha": alpha})
                self.set_artist_property(a, lpp)

            self.delp("loaded_property")

        dprint3("generating axspan artist")
        for artist in self._artists:
            artist.figobj = self
            artist.figobj_hl = []
            artist.set_zorder(self.getp('zorder'))
#             self._objs.append(weakref.ref(artist))

    def del_artist(self, artist=None, delall=False):
        #        if delall:
        #           artistlist=self._artists
        #        else:
        #           artistlist=artist
        artistlist = self._artists
        self.store_loaded_property()

#        val = []
#        for a in self._artists:
#           val.append(self.get_artist_property(a))
#        if len(val) != 0:
#           self.setp("loaded_property", val)

        if len(artistlist) != 0:
            self.highlight_artist(False, artistlist)
            container = self.get_container()
            for a in artistlist:
                a.set_picker(None)
                a.remove()
#             container.lines.remove(a)

        super(FigAxspan, self).del_artist(artist=artistlist)

    def highlight_artist(self, val, artist=None):
        if artist is None:
            alist = self._artists
        else:
            alist = artist
#        container=self._parent._artists[0]
        container = self.get_container()
        if val == True:
            hl = []
            x, y = self._eval_xy()
            for x1 in x:
                hl.append(container.axvline(x1[0],  marker='s',
                                            color='k', linestyle='None',
                                            markeredgewidth=0.5,
                                            markerfacecolor='None'))
                hl.append(container.axvline(x1[1],  marker='s',
                                            color='k', linestyle='None',
                                            markeredgewidth=0.5,
                                            markerfacecolor='None'))
            for y1 in y:
                hl.append(container.axhline(y1[0],  marker='s',
                                            color='k', linestyle='None',
                                            markeredgewidth=0.5,
                                            markerfacecolor='None'))
                hl.append(container.axhline(y1[1],  marker='s',
                                            color='k', linestyle='None',
                                            markeredgewidth=0.5,
                                            markerfacecolor='None'))
            for a in alist:
                a.figobj_hl += hl

        else:
            for a in self._artists:
                for hl in a.figobj_hl:
                    try:
                        hl.remove()
                    except:
                        pass
                a.figobj_hl = []

#
#   HitTest
#
    def picker_a(self, a, evt):
        ###
        # it does not use contain to catch edges
        ###
        if not a in self._artists:
            return False, {}

        idx = self._artists.index(a)
        x = self.getp("x")
        if idx >= len(x):
            mode = 'h'
        else:
            mode = 'v'
        xy = a.get_xy()
        axes = a.axes
        x0d, y0d = transform_point(axes.transData,
                                   min(xy[:, 0]), min(xy[:, 1]))
        x1d, y1d = transform_point(axes.transData,
                                   max(xy[:, 0]), max(xy[:, 1]))
        x0da, y0da = transform_point(axes.transAxes,
                                     min(xy[:, 0]), min(xy[:, 1]))
        x1da, y1da = transform_point(axes.transAxes,
                                     max(xy[:, 0]), max(xy[:, 1]))

        hit = False
        if mode == 'v':
            if (evt.x > x0d-5 and
                evt.x < x1d+5 and
                evt.y > y0da and
                    evt.y < y1da):
                hit = True
        else:
            if (evt.x > x0da and
                evt.x < x1da and
                evt.y > y0d-5 and
                    evt.y < y1d+5):
                hit = True
        if hit:
            return True, {'child_artist': a}
        return False, {}

    def picker_a0(self, artist, evt):
        hit, extra = self.picker_a(artist, evt)
        if hit:
            return hit, extra, 'area', 3
        else:
            return False, {}, None, 0

    def dragstart_a(self, a, evt):
        return self.dragstart(a, evt)

    def dragstart(self, a, evt):
        if evt.inaxes is None:
            return 0
        axes = a.axes
        if axes is None:
            return 0

        if not a in self._artists:
            return False, {}
        idx = self._artists.index(a)
        x = self.getp("x")
        if idx >= len(x):
            mode = 'h'
        else:
            mode = 'v'

        self._drag_backup = (evt.xdata, evt.ydata, evt.x, evt.y)
        self._drag_delta = (0, 0)
        self._drag_mode = 1
        # 1 transpose
        # 2 x expand(smaller edge)
        # 3 x expand(larger edge)
        # 4 y expand(smaller edge)
        # 5 y expand(smaller edge)
        xy = a.get_xy()
        x0d, y0d = transform_point(axes.transData,
                                   min(xy[:, 0]), min(xy[:, 1]))
        x1d, y1d = transform_point(axes.transData,
                                   max(xy[:, 0]), max(xy[:, 1]))
#        print evt.x, x0d

        delta = 3
        if mode == 'v':
            if abs(evt.x - x0d) < delta:
                self._drag_mode = 2
            if abs(evt.x - x1d) < delta:
                self._drag_mode = 3
        else:
            if abs(evt.y - y0d) < delta:
                self._drag_mode = 4
            if abs(evt.y - y1d) < delta:
                self._drag_mode = 5

        return 0

    def drag_a(self, a, evt, shift=None, scale=None):
        return self.drag(a, evt), scale

    def drag(self, a, evt, idx='all'):
        if evt.xdata is None:
            return
        if evt.inaxes is None:
            return
        axes = a.axes
        x, y = self._eval_xy()
        dx = evt.xdata - self._drag_backup[0]
        dy = evt.ydata - self._drag_backup[1]
        dxd = (evt.x - self._drag_backup[2])
        dyd = (evt.y - self._drag_backup[3])
        if (evt.guiEvent.ShiftDown()):
            if abs(dxd) > 2*abs(dyd):
                dyd = 0
            elif abs(dyd) > 2*abs(dxd):
                dxd = 0

        i = 0
        for x0 in x:
            if idx != 'all' and idx != i:
                i = i+1
                continue
            xd, void = transform_point(axes.transData,
                                       x0[0], evt.ydata)
            x1, void = transform_point(axes.transData.inverted(),
                                       xd+dxd, void)
            xd, void = transform_point(axes.transData,
                                       x0[1], evt.ydata)
            x2, void = transform_point(axes.transData.inverted(),
                                       xd+dxd, void)
            if self._drag_mode == 1:
                xx = [x1, x2]
            elif self._drag_mode == 2:
                xx = [x1, x0[1]]
            elif self._drag_mode == 3:
                xx = [x0[0], x2]
            elif self._drag_mode == 4:
                xx = x0
            elif self._drag_mode == 5:
                xx = x0
            xy = _make_xy_x(xx)
            self._artists[i].set_xy(xy)
            i = i+1
        for y0 in y:
            if idx != 'all' and idx != i:
                i = i+1
                continue
            void, yd = transform_point(axes.transData,
                                       evt.xdata, y0[0])
            void, y1 = transform_point(axes.transData.inverted(),
                                       void, yd+dyd)
            void, yd = transform_point(axes.transData,
                                       evt.xdata, y0[1])
            void, y2 = transform_point(axes.transData.inverted(),
                                       void, yd+dyd)
            if self._drag_mode == 1:
                yy = [y1, y2]
            elif self._drag_mode == 2:
                yy = y0
            elif self._drag_mode == 3:
                yy = y0
            elif self._drag_mode == 4:
                yy = [y1, y0[1]]
            elif self._drag_mode == 5:
                yy = [y0[0], y2]
            xy = _make_xy_y(yy)
            self._artists[i].set_xy(xy)
            i = i+1
        self._drag_delta = (dxd, dyd)

    def drag_a_get_hl(self, a):
        return self.drag_get_hl(a)

    def drag_get_hl(self, a):
        self._alpha_backup = self.get_alpha(None)
        self.set_alpha(0.5, None)
        return self._artists

    def drag_a_rm_hl(self, a):
        return self.drag_rm_hl(a)

    def drag_rm_hl(self, a):
        self.set_alpha(self._alpha_backup, None)

    def dragdone_a(self, a, evt, shift=None, scale=None):
        return self.dragdone(a, evt), scale

    def dragdone(self, a, evt, idx='all'):
        axes = a.axes
        x, y = self._eval_xy()
        x = x.copy()
        y = y.copy()
        dxd, dyd = self._drag_delta

        if (evt.xdata is not None and
                evt.ydata is not None):
            i = 0
            xdata = []
            ydata = []
            for x0 in x:
                x1 = x0[:]
                if idx == 'all' or idx == i:
                    if (self._drag_mode == 1 or
                            self._drag_mode == 2):
                        xd, void = transform_point(axes.transData,
                                                   x0[0], evt.ydata)
                        x1[0], void = transform_point(axes.transData.inverted(),
                                                      xd+dxd, void)
                    if (self._drag_mode == 1 or
                            self._drag_mode == 3):
                        xd, void = transform_point(axes.transData,
                                                   x0[1], evt.ydata)
                        x1[1], void = transform_point(axes.transData.inverted(),
                                                      xd+dxd, void)
                xdata.append(x1)
                i = i + 1
            for y0 in y:
                y1 = y0[:]
                if idx == 'all' or idx == i:
                    if (self._drag_mode == 1 or
                            self._drag_mode == 4):
                        void, yd = transform_point(axes.transData,
                                                   evt.xdata, y0[0])
                        void, y1[0] = transform_point(axes.transData.inverted(),
                                                      void, yd+dyd)
                    if (self._drag_mode == 1 or
                            self._drag_mode == 5):
                        void, yd = transform_point(axes.transData,
                                                   evt.xdata, y0[1])
                        void, y1[1] = transform_point(axes.transData.inverted(),
                                                      void, yd+dyd)
                ydata.append(y1)
                i = i + 1
        else:
            xdata = x
            ydata = y

        xdata = np.array(xdata)
        ydata = np.array(ydata)
        action = UndoRedoFigobjMethod(self._artists[0],
                                      'data', (xdata, ydata))

        window = evt.guiEvent_memory.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry([action])
        return 1
#
#  Popup in Canvas
#

    def onExport(self, event):
        from matplotlib.artist import getp
        canvas = event.GetEventObject()
        sel = [a() for a in canvas.selection]
        shell = self.get_root_parent().app.shell
        for a in self._artists:
            if a in sel:
                print("Exporting Data to Shell")
                xy = a.get_xy()
                fig_val = {"xdata": (min(xy[:, 0]), max(xy[:, 0])),
                           "ydata": (min(xy[:, 1]), max(xy[:, 1])), }
                self.write2shell(fig_val, "fig_val")
                break
        shell.redirectStdout(True)
        text = '#Exporting data as fig_val[\'xdata\'], fig_val[\'ydata\']\"'
        shell.writeOut(text)
        shell.redirectStdout(False)
        # print "exporting data to file"

#
#   setter/getter
#
    def set_data(self, value, a):
        if (len(self._artists) != len(value[0]) + len(value[1]) or
            not _is_data_clean(value[0]) or
                not _is_data_clean(value[1])):
            x = _clean_data(value[0])
            y = _clean_data(value[1])
            x = _sort_data(x)
            y = _sort_data(y)
            self.setp("x", np.array(x))
            self.setp("y", np.array(y))
            self.del_artist()
            self.generate_artist()
        else:
            x = _sort_data(value[0])
            y = _sort_data(value[1])
            self.setp("x", np.array(x))
            self.setp("y", np.array(y))
            i = 0
            for x0 in x:
                xy = _make_xy_x(x0)
                self._artists[i].set_xy(xy)
                i = i+1
            for y0 in y:
                xy = _make_xy_y(y0)
                self._artists[i].set_xy(xy)
                i = i+1
        self._data_extent = None

    def get_data(self, a):
        return (self.getp('x')[:], self.getp('y')[:])

    def set_alpha(self, value, a):
        set_mpl_all(self._artists, 'alpha', value)

    def get_alpha(self, a):
        return get_mpl_first(self._artists, 'alpha')

    def set_fill(self, value, a):
        set_mpl_all(self._artists, 'fill', value)

    def get_fill(self, a):
        return get_mpl_first(self._artists, 'fill')

    def set_facecolor(self, value, a):
        set_mpl_all(self._artists, 'facecolor', value)

    def get_facecolor(self, a):
        return get_mpl_first(self._artists, 'facecolor')

    def set_edgecolor(self, value, a):
        set_mpl_all(self._artists, 'edgecolor', value)

    def get_edgecolor(self, a):
        return get_mpl_first(self._artists, 'edgecolor')

    def set_linestyle(self, value, a):
        set_mpl_all(self._artists, 'linestyle', value)

    def get_linestyle(self, a):
        return get_mpl_first(self._artists, 'linestyle')

    def set_linewidth(self, value, a):
        set_mpl_all(self._artists, 'linewidth', value)

    def get_linewidth(self, a):
        return get_mpl_first(self._artists, 'linewidth')

#
#   data extent
#
    def get_data_extent(self):
        if self._data_extent is not None:
            return self._data_extent
        x, y = self._eval_xy()
        d = [None]*4
        if len(x) != 0:
            x1 = [min(item) for item in x if len(item) == 2]
            x2 = [max(item) for item in x if len(item) == 2]
            d[0] = min(x1)
            d[1] = max(x2)
        if len(y) != 0:
            y1 = [min(item) for item in y if len(item) == 2]
            y2 = [max(item) for item in y if len(item) == 2]
            d[2] = min(y1)
            d[3] = max(y2)
        self._data_extent = d
        return self._data_extent

#
#   range setting
#
    def get_xrange(self, xrange=[None, None], scale='linear'):
        x, y = self._eval_xy()
        if scale == 'log':
            x = mask_negative(x)
        if x.size > 0:
            return self._update_range(xrange, [np.nanmin(x), np.nanmax(x)])
        else:
            return super(FigAxspan, self).get_xrange(xrange=xrange, scale=scale)

    def get_yrange(self, yrange=[None, None], xrange=[None, None], scale='linear'):
        x, y = self._eval_xy()
        if scale == 'log':
            y = mask_negative(y)
        if y.size != 0:
            return self._update_range(yrange, [np.nanmin(y), np.nanmax(y)])
        else:
            return super(FigAxspan, self).get_yrange(yrange=yrange,
                                                     xrange=xrange, scale=scale)

    def save_data2(self, data=None):
        def check(obj, name):
            return obj.getp(name) is obj.getvar(name)
        if data is None:
            data = {}
        var = {'x': check(self, 'x'),
               'y': check(self, 'y')}

        if var["x"] is not True:
            var["xdata"] = self.getp("x")
        if var["y"] is not True:
            var["ydata"] = self.getp("y")

        data['FigAxspan'] = (1, var)
        data = super(FigAxspan, self).save_data2(data)
        return data

    def load_data2(self, data):
        d = data['FigAxspan']
        super(FigAxspan, self).load_data2(data)
        var = d[1]
        names = ["x", "y"]
        for name in names:
            if var[name]:
                self.setp(name,
                          self.getvar(name))
            else:
                self.setp(name, var[name+'data'])

    def _eval_xy(self):
        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return None, None
        return self.getp(("x", "y"))



from __future__ import print_function
from ifigure.utils.edit_list import DialogEditListTab
import ifigure.widgets.dialog as dialog
import wx
pkeys = ['experiment', 'default_node', 'x', 'y', 'z',
         'xerr', 'yerr', 'title']


def MDS_setting(parent, setting,  gwparam, mds_server, thread_main):
    def make_gs_arr(gs, s):
        return [gs[key][s] for key in pkeys]

    gs = gwparam.getvar('globalsets')
    gv = gwparam.getvar()
    tab = ['Set1', 'Set2', 'Set3', 'Set4', 'config', 'more']

    list1 = [[None, make_gs_arr(gs, 0), 118, None]]
    list2 = [[None, make_gs_arr(gs, 1), 118, None]]
    list3 = [[None, make_gs_arr(gs, 2), 118, None]]
    list4 = [[None, make_gs_arr(gs, 3), 118, None]]

    txt = "on" if parent.mpanalysis else "off"
    txt2 = "on" if gv['use_shot_global'] else "off"
    txt7 = "on" if parent.init_beginning else "off"
    txt6 = "on" if parent.init_sec else "off"
    txt5 = "on" if parent.init_page else "off"

    list5 = [["MDSconnect", mds_server, 204, None],
             ["Update", gv['event'], 0, None],
             ["Common vars", gv['shot_global'], 0, None],
             ["Use common vars",  txt2,  1, {"values": ["on", "off"]}],
             ["Global setting", gv['global_choice'], 21, None],
             ["Multipage mode",  txt, 1, {"values": ["on", "off"]}],
             #            ["Color Order", gv['color_order'], 22, {}],]
             ]

    tmp2 = [[None,  parent.parallel_sec, 3,
             {"text": 'Parallelize Panels', "noindent": True}], ]
    tmp = [[None,  (parent.parallel_shot, [parent.parallel_sec]), 27,
            ({"text": 'Parallelize Pages'}, {"elp": tmp2})]]
    list6 = [[None, "MDSplus session parallelization",  102, None],
             [None,  (parent.parallel_page,
                      [(parent.parallel_shot, [parent.parallel_sec])]),
              27,
              ({"text": 'Parallelize Shots', "space": 15}, {"elp": tmp})],
             #["At session start",txt7, 1, {"values":["on", "off"]}],
             #["At every page",   txt5, 1, {"values":["on", "off"]}],
             #["At every panel",  txt6, 1, {"values":["on", "off"]}],
             [None, thread_main,  3,
              {"text": "Run python script in main thread",
               "noindent": None}],
             [None, "(!) MDS+ connection is initialized when opening new tree",  102, None], ]

    tip1 = None
    tip2 = None
    tip3 = None
    tip4 = None
    tip5 = ['MDSplus server and default Tree',
            'Update event',
            'MDS command called at the begining of page',
            'MDS command called at the begining of section',
            'Global setting selection for shots',
            'Use the same shot numbers for all pages. (shot numbers after semi-colon are ignored)', ]

    tip6 = [None,
            'Parallelization of MDS session',
            None,
            #'Initialize global/local variables when analysis loop starts',
            #'Initialize global/local variables when analyzing a new panel',
            #'Initialize global/local variables when analyzing a new page',
            None, ]

    l = [list1, list2, list3, list4, list5, list6]
    tip = [tip1, tip2, tip3, tip4, tip5, tip6]

    value = DialogEditListTab(tab, l, modal=False, tip=tip, parent=parent,
                              title='Global Setting',
                              style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

    if value[0]:
        for m in range(4):
            for i, v in enumerate(value[1][m][0]):
                gs[pkeys[i]][m] = str(v[1])
        gv['event'] = str(value[1][4][1])
        gv['shot_global'] = str(value[1][4][2])
        gv['use_shot_global'] = (value[1][4][3] == 'on')
        gv['global_choice'] = value[1][4][4]
        parent.mpanalysis = (value[1][4][5] == "on")
#      gv['color_order'] = value[1][4][5]

#      parent.init_beginning = (value[1][5][3] == "on")
#      parent.init_page = (value[1][5][4] == "on")
#      parent.init_sec = (value[1][5][5] == "on")

        if not value[1][5][1][0]:
            parent.parallel_page = False
            parent.parallel_shot = False
            parent.parallel_sec = False
        else:
            parent.parallel_page = True
            if not value[1][5][1][1][0][0]:
                parent.parallel_shot = False
                parent.parallel_sec = False
            else:
                parent.parallel_shot = True
                parent.parallel_sec = value[1][5][1][1][0][1][0]

        parent.book.setvar('mdsscript_main',
                           value[1][5][2])
        parent.book.setvar('mdsplus_server',
                           str(value[1][4][0]))

    from ifigure.utils.setting_parser import iFigureSettingParser as SettingParser
    p = SettingParser()
    v = p.read_setting('mdsplus.default_mdsserver')
    if str(v['server']) != str(value[1][4][0]):
        ret = dialog.message(parent,
                             '\n'.join(['Selected MDS+ server is different from your default setting',
                                        'Do you want to use this server as your default server?']),
                             'MDS+ server setting',
                             2, icon=wx.ICON_QUESTION)
        print(ret)
        if ret == 'ok':
            v['server'] = str(value[1][4][0])
            p.write_setting('mdsplus.default_mdsserver', v)

    return value


def pref_setting(parent, setting):
    pass



from __future__ import print_function
# {{{ http://code.activestate.com/recipes/576704/ (r16)
#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       pyminifier.py
#
#       Copyright 2009 Dan McDougall <YouKnowWho@YouKnowWhat.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; Version 3 of the License
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, the license can be downloaded here:
#
#       http://www.gnu.org/licenses/gpl.html

# Meta
__version__ = '1.4.1'
__license__ = "GNU General Public License (GPL) Version 3"
__version_info__ = (1, 4, 1)
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

"""
**Python Minifier:**  Reduces the size of (minifies) Python code for use on
embedded platforms.

Performs the following:
     - Removes docstrings.
     - Removes comments.
     - Minimizes code indentation.
     - Joins multiline pairs of parentheses, braces, and brackets (and removes extraneous whitespace within).
     - Preserves shebangs and encoding info (e.g. "# -- coding: utf-8 --").

Various examples and edge cases are sprinkled throughout the pyminifier code so
that it can be tested by minifying itself.  The way to test is thus:

.. code-block:: bash

    $ python pyminifier.py pyminifier.py > minified_pyminifier.py
    $ python minified_pyminifier.py pyminifier.py > this_should_be_identical.py
    $ diff minified_pyminifier.py this_should_be_identical.py
    $

If you get an error executing minified_pyminifier.py or
'this_should_be_identical.py' isn't identical to minified_pyminifier.py then
something is broken.
"""

import sys
import re
from six.moves import cStringIO as StringIO
import tokenize
from optparse import OptionParser

# Compile our regular expressions for speed
multiline_quoted_string = re.compile(r'(\'\'\'|\"\"\")')
not_quoted_string = re.compile(r'(\".*\'\'\'.*\"|\'.*\"\"\".*\')')
trailing_newlines = re.compile(r'\n\n')
shebang = re.compile(r'^#\!.*$')
encoding = re.compile(r".*coding[:=]\s*([-\w.]+)")
multiline_indicator = re.compile(r'\\\\(\s*#.*)?\n')
# The above also removes trailing comments: "test = 'blah \ # comment here"

# These aren't used but they're a pretty good reference:
double_quoted_string = re.compile(r'((?<!\\)".*?(?<!\\)")')
single_quoted_string = re.compile(r"((?<!\\)'.*?(?<!\\)')")
single_line_single_quoted_string = re.compile(r"((?<!\\)'''.*?(?<!\\)''')")
single_line_double_quoted_string = re.compile(r"((?<!\\)'''.*?(?<!\\)''')")


def remove_comments_and_docstrings(source):
    """
    Returns 'source' minus comments and docstrings.

    **Note**: Uses Python's built-in tokenize module to great effect.

    Example:

    .. code-block:: python

        def noop(): # This is a comment
            '''
            Does nothing.
            '''
            pass # Don't do anything

    Will become:

    .. code-block:: python

        def noop():
            pass
    """
    io_obj = StringIO(source)
    out = ""
    prev_toktype = tokenize.INDENT
    last_lineno = -1
    last_col = 0
    for tok in tokenize.generate_tokens(io_obj.readline):
        token_type = tok[0]
        token_string = tok[1]
        start_line, start_col = tok[2]
        end_line, end_col = tok[3]
        ltext = tok[4]
        # The following two conditionals preserve indentation.
        # This is necessary because we're not using tokenize.untokenize()
        # (because it spits out code with copious amounts of oddly-placed
        # whitespace).
        if start_line > last_lineno:
            last_col = 0
        if start_col > last_col:
            out += (" " * (start_col - last_col))
        # Remove comments:
        if token_type == tokenize.COMMENT:
            pass
        # This series of conditionals removes docstrings:
        elif token_type == tokenize.STRING:
            if prev_toktype != tokenize.INDENT:
                # This is likely a docstring; double-check we're not inside an operator:
                if prev_toktype != tokenize.NEWLINE:
                    # Note regarding NEWLINE vs NL: The tokenize module
                    # differentiates between newlines that start a new statement
                    # and newlines inside of operators such as parens, brackes,
                    # and curly braces.  Newlines inside of operators are
                    # NEWLINE and newlines that start new code are NL.
                    # Catch whole-module docstrings:
                    if start_col > 0:
                        # Unlabelled indentation means we're inside an operator
                        out += token_string
                    # Note regarding the INDENT token: The tokenize module does
                    # not label indentation inside of an operator (parens,
                    # brackets, and curly braces) as actual indentation.
                    # For example:
                    # def foo():
                    #     "The spaces before this docstring are tokenize.INDENT"
                    #     test = [
                    #         "The spaces before this string do not get a token"
                    #     ]
        else:
            out += token_string
        prev_toktype = token_type
        last_col = end_col
        last_lineno = end_line
    return out


def reduce_operators(source):
    """
    Remove spaces between operators in 'source' and returns the result.

    Example:

    .. code-block:: python

        def foo(foo, bar, blah):
            test = "This is a %s" % foo

    Will become:

    .. code-block:: python

        def foo(foo,bar,blah):
            test="This is a %s"%foo
    """
    io_obj = StringIO(source)
    remove_columns = []
    out = ""
    out_line = ""
    prev_toktype = tokenize.INDENT
    prev_tok = None
    last_lineno = -1
    last_col = 0
    lshift = 1
    for tok in tokenize.generate_tokens(io_obj.readline):
        token_type = tok[0]
        token_string = tok[1]
        start_line, start_col = tok[2]
        end_line, end_col = tok[3]
        ltext = tok[4]
        if start_line > last_lineno:
            last_col = 0
        if start_col > last_col:
            out_line += (" " * (start_col - last_col))
        if token_type == tokenize.OP:
            # Operators that begin a line such as @ or open parens should be
            # left alone
            start_of_line_types = [  # These indicate we're starting a new line
                tokenize.NEWLINE, tokenize.DEDENT, tokenize.INDENT]
            if prev_toktype not in start_of_line_types:
                # This is just a regular operator; remove spaces
                remove_columns.append(start_col)  # Before OP
                remove_columns.append(end_col+1)  # After OP
        if token_string.endswith('\n'):
            out_line += token_string
            if remove_columns:
                for col in remove_columns:
                    col = col - lshift
                    try:
                        # This was really handy for debugging (looks nice, worth saving):
                        # print out_line + (" " * col) + "^"
                        # The above points to the character we're looking at
                        if out_line[col] == " ":  # Only if it is a space
                            out_line = out_line[:col] + out_line[col+1:]
                            lshift += 1  # To re-align future changes on this line
                    except IndexError:  # Reached and end of line, no biggie
                        pass
            out += out_line
            remove_columns = []
            out_line = ""
            lshift = 1
        else:
            out_line += token_string
        prev_toktype = token_type
        prev_token = tok
        last_col = end_col
        last_lineno = end_line
    # This makes sure to capture the last line if it doesn't end in a newline:
    out += out_line
    # The tokenize module doesn't recognize @ sign before a decorator
    return out

# NOTE: This isn't used anymore...  Just here for reference in case someone
# searches the internet looking for a way to remove similarly-styled end-of-line
# comments from non-python code.  It also acts as an edge case of sorts with
# that raw triple quoted string inside the "quoted_string" assignment.


def remove_comment(single_line):
    """
    Removes the comment at the end of the line (if any) and returns the result.
    """
    quoted_string = re.compile(
        r'''((?<!\\)".*?(?<!\\)")|((?<!\\)'.*?(?<!\\)')'''
    )
    # This divides the line up into sections:
    #   Those inside single quotes and those that are not
    split_line = quoted_string.split(single_line)
    # Remove empty items:
    split_line = [a for a in split_line if a]
    out_line = ""
    for section in split_line:
        if section.startswith("'") or section.startswith('"'):
            # This is a quoted string; leave it alone
            out_line += section
        elif '#' in section:  # A '#' not in quotes?  There's a comment here!
            # Get rid of everything after the # including the # itself:
            out_line += section.split('#')[0]
            break  # No reason to bother the rest--it's all comments
        else:
            # This isn't a quoted string OR a comment; leave it as-is
            out_line += section
    return out_line.rstrip()  # Strip trailing whitespace before returning


def join_multiline_pairs(text, pair="()"):
    """
    Finds and removes newlines in multiline matching pairs of characters in
    'text'.  For example, "(.*\n.*), {.*\n.*}, or [.*\n.*]".

    By default it joins parens () but it will join any two characters given via
    the 'pair' variable.

    **Note:** Doesn't remove extraneous whitespace that ends up between the pair.
    Use reduce_operators() for that.

    Example:

    .. code-block:: python

        test = (
            "This is inside a multi-line pair of parentheses"
        )

    Will become:

    .. code-block:: python

        test = (            "This is inside a multi-line pair of parentheses"        )
    """
    # Readability variables
    opener = pair[0]
    closer = pair[1]

    # Tracking variables
    inside_pair = False
    inside_quotes = False
    inside_double_quotes = False
    inside_single_quotes = False
    quoted_string = False
    openers = 0
    closers = 0
    linecount = 0

    # Regular expressions
    opener_regex = re.compile(r'\%s' % opener)
    closer_regex = re.compile(r'\%s' % closer)

    output = ""

    for line in text.split('\n'):
        escaped = False
        # First we rule out multi-line strings
        multline_match = multiline_quoted_string.search(line)
        not_quoted_string_match = not_quoted_string.search(line)
        if multline_match and not not_quoted_string_match and not quoted_string:
            if len(line.split('"""')) > 1 or len(line.split("'''")):
                # This is a single line that uses the triple quotes twice
                # Treat it as if it were just a regular line:
                output += line + '\n'
                quoted_string = False
            else:
                output += line + '\n'
                quoted_string = True
        elif quoted_string and multiline_quoted_string.search(line):
            output += line + '\n'
            quoted_string = False
        # Now let's focus on the lines containing our opener and/or closer:
        elif not quoted_string:
            if opener_regex.search(line) or closer_regex.search(line) or inside_pair:
                for character in line:
                    if character == opener:
                        if not escaped and not inside_quotes:
                            openers += 1
                            inside_pair = True
                            output += character
                        else:
                            escaped = False
                            output += character
                    elif character == closer:
                        if not escaped and not inside_quotes:
                            if openers and openers == (closers + 1):
                                closers = 0
                                openers = 0
                                inside_pair = False
                                output += character
                            else:
                                closers += 1
                                output += character
                        else:
                            escaped = False
                            output += character
                    elif character == '\\':
                        if escaped:
                            escaped = False
                            output += character
                        else:
                            escaped = True
                            output += character
                    elif character == '"' and escaped:
                        output += character
                        escaped = False
                    elif character == "'" and escaped:
                        output += character
                        escaped = False
                    elif character == '"' and inside_quotes:
                        if inside_single_quotes:
                            output += character
                        else:
                            inside_quotes = False
                            inside_double_quotes = False
                            output += character
                    elif character == "'" and inside_quotes:
                        if inside_double_quotes:
                            output += character
                        else:
                            inside_quotes = False
                            inside_single_quotes = False
                            output += character
                    elif character == '"' and not inside_quotes:
                        inside_quotes = True
                        inside_double_quotes = True
                        output += character
                    elif character == "'" and not inside_quotes:
                        inside_quotes = True
                        inside_single_quotes = True
                        output += character
                    elif character == ' ' and inside_pair and not inside_quotes:
                        if not output[-1] in [' ', opener]:
                            output += ' '
                    else:
                        if escaped:
                            escaped = False
                        output += character
                if inside_pair == False:
                    output += '\n'
            else:
                output += line + '\n'
        else:
            output += line + '\n'

    # Clean up
    output = trailing_newlines.sub('\n', output)

    return output


def dedent(source):
    """
    Minimizes indentation to save precious bytes

    Example:

    .. code-block:: python

        def foo(bar):
            test = "This is a test"

    Will become:

    .. code-block:: python

        def foo(bar):
         test = "This is a test"
    """
    io_obj = StringIO(source)
    out = ""
    last_lineno = -1
    last_col = 0
    prev_start_line = 0
    indentation = ""
    indentation_level = 0
    for i, tok in enumerate(tokenize.generate_tokens(io_obj.readline)):
        token_type = tok[0]
        token_string = tok[1]
        start_line, start_col = tok[2]
        end_line, end_col = tok[3]
        if start_line > last_lineno:
            last_col = 0
        if token_type == tokenize.INDENT:
            indentation_level += 1
            continue
        if token_type == tokenize.DEDENT:
            indentation_level -= 1
            continue
        indentation = " " * indentation_level
        if start_line > prev_start_line:
            out += indentation + token_string
        elif start_col > last_col:
            out += " " + token_string
        else:
            out += token_string
        prev_start_line = start_line
        last_col = end_col
        last_lineno = end_line
    return out


def fix_empty_methods(source):
    """
    Appends 'pass' to empty methods/functions (i.e. where there was nothing but
    a docstring before we removed it =).

    Example:

    .. code-block:: python

        # Note: This triple-single-quote inside a triple-double-quote is also a
        # pyminifier self-test
        def myfunc():
            '''This is just a placeholder function.'''

    Will become:

    .. code-block:: python

        def myfunc(): pass
    """
    def_indentation_level = 0
    output = ""
    just_matched = False
    previous_line = None
    method = re.compile(r'^\s*def\s*.*\(.*\):.*$')
    for line in source.split('\n'):
        if len(line.strip()) > 0:  # Don't look at blank lines
            if just_matched == True:
                this_indentation_level = len(line.rstrip()) - len(line.strip())
                if def_indentation_level == this_indentation_level:
                    # This method is empty, insert a 'pass' statement
                    output += "%s pass\n%s\n" % (previous_line, line)
                else:
                    output += "%s\n%s\n" % (previous_line, line)
                just_matched = False
            elif method.match(line):
                def_indentation_level = len(
                    line) - len(line.strip())  # A commment
                just_matched = True
                previous_line = line
            else:
                output += "%s\n" % line  # Another self-test
        else:
            output += "\n"
    return output


def remove_blank_lines(source):
    """
    Removes blank lines from 'source' and returns the result.

    Example:

    .. code-block:: python

        test = "foo"

        test2 = "bar"

    Will become:

    .. code-block:: python

        test = "foo"
        test2 = "bar"
    """
    io_obj = StringIO(source)
    source = [a for a in io_obj.readlines() if a.strip()]
    return "".join(source)


def minify(source):
    """
    Remove all docstrings, comments, blank lines, and minimize code
    indentation from 'source' then prints the result.
    """
    preserved_shebang = None
    preserved_encoding = None

    # This is for things like shebangs that must be precisely preserved
    for line in source.split('\n')[0:2]:
        # Save the first comment line if it starts with a shebang
        # (e.g. '#!/usr/bin/env python') <--also a self test!
        if shebang.match(line):  # Must be first line
            preserved_shebang = line
            continue
        # Save the encoding string (must be first or second line in file)
        if encoding.match(line):
            preserved_encoding = line

    # Remove multilines (e.g. lines that end with '\' followed by a newline)
    source = multiline_indicator.sub('', source)

    # Remove docstrings (Note: Must run before fix_empty_methods())
    source = remove_comments_and_docstrings(source)

    # Remove empty (i.e. single line) methods/functions
    source = fix_empty_methods(source)

    # Join multiline pairs of parens, brackets, and braces
    source = join_multiline_pairs(source)
    source = join_multiline_pairs(source, '[]')
    source = join_multiline_pairs(source, '{}')

    # Remove whitespace between operators:
    source = reduce_operators(source)

    # Minimize indentation
    source = dedent(source)

    # Re-add preseved items
    if preserved_encoding:
        source = preserved_encoding + "\n" + source
    if preserved_shebang:
        source = preserved_shebang + "\n" + source

    # Remove blank lines
    source = remove_blank_lines(source).rstrip('\n')  # Stubborn last newline

    return source


def bz2_pack(source):
    "Returns 'source' as a bzip2-compressed, self-extracting python script."
    import bz2
    import base64
    out = ""
    compressed_source = bz2.compress(source)
    out += 'import bz2, base64\n'
    out += "exec bz2.decompress(base64.b64decode('"
    out += base64.b64encode((compressed_source))
    out += "'))\n"
    return out


def gz_pack(source):
    "Returns 'source' as a gzip-compressed, self-extracting python script."
    import zlib
    import base64
    out = ""
    compressed_source = zlib.compress(source)
    out += 'import zlib, base64\n'
    out += "exec zlib.decompress(base64.b64decode('"
    out += base64.b64encode((compressed_source))
    out += "'))\n"
    return out

# The test.+() functions below are for testing pyminifer...


def test_decorator(f):
    """Decorator that does nothing"""
    return f


def test_reduce_operators():
    """Test the case where an operator such as an open paren starts a line"""
    (a, b) = 1, 2  # The indentation level should be preserved
    pass


def test_empty_functions():
    """
    This is a test method.
    This should be replaced with 'def empty_method: pass'
    """


class test_class(object):
    "Testing indented decorators"

    @test_decorator
    def foo(self):
        pass


def test_function():
    """
    This function encapsulates the edge cases to prevent them from invading the
    global namespace.
    """
    foo = ("The # character in this string should "  # This comment
           "not result in a syntax error")  # ...and this one should go away
    test_multi_line_list = [
        'item1',
        'item2',
        'item3'
    ]
    test_multi_line_dict = {
        'item1': 1,
        'item2': 2,
        'item3': 3
    }
    # It may seem strange but the code below tests our docstring removal code.
    test_string_inside_operators = imaginary_function(
        "This string was indented but the tokenizer won't see it that way."
    )  # To understand how this could mess up docstring removal code see the
    # remove_comments_and_docstrings() function starting at this line:
    #     "elif token_type == tokenize.STRING:"
    # This tests remove_extraneous_spaces():
    this_line_has_leading_indentation = '''<--That extraneous space should be
                                              removed'''  # But not these spaces


def main():
    usage = '%prog [options] "<input file>"'
    parser = OptionParser(usage=usage, version=__version__)
    parser.disable_interspersed_args()
    parser.add_option(
        "-o", "--outfile",
        dest="outfile",
        default=None,
        help="Save output to the given file.",
        metavar="<file path>"
    )
    parser.add_option(
        "--bzip2",
        action="store_true",
        dest="bzip2",
        default=False,
        help="bzip2-compress the result into a self-executing python script."
    )
    parser.add_option(
        "--gzip",
        action="store_true",
        dest="gzip",
        default=False,
        help="gzip-compress the result into a self-executing python script."
    )
    options, args = parser.parse_args()
    try:
        source = open(args[0]).read()
    except Exception as e:
        print(e)
        parser.print_help()
        sys.exit(2)
    # Minify our input script
    result = minify(source)
    # Compress it if we were asked to do so
    if options.bzip2:
        result = bz2_pack(result)
    elif options.gzip:
        result = gz_pack(result)
    # Either save the result to the output file or print it to stdout
    if options.outfile:
        f = open(options.outfile, 'w')
        f.write(result)
        f.close()
    else:
        print(result)


if __name__ == "__main__":
    main()
# end of http://code.activestate.com/recipes/576704/ }}}



from ifigure.mto.treedict import TreeDict


class PyClass(TreeDict):
    """ 
    A tree object class which all the essential 
    capabilities are defined externally
    """

    def __init__(self, parent, src=None):
        self._module_file = ''      # file name
        self._module_co = None
        self._module_mtime = 0.
        super(PyClass, self).__init__(parent=parent, src=src)

    @classmethod
    def isPyClass(self):
        return True

    def get_namebase(self):
        return 'classmodule'

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'script.png')
        return [idx1]

    def run_init(self, e=None):
        self._module.run_init(self)

    def do_run(self, e=None):
        pass

    def popup_menu(self):
        pass

    def load_module(self, file=None):
        if file is None:
            self._module_co = None

        logging.basicConfig(level=logging.DEBUG)
        try:
            self._module_co = imp.load_source('ifigure.add_on.tmp', file)
            name = self._module_co.module_name
            self._module_co = imp.load_source('ifigure.add_on'+name, file)
            self._module_mtime = os.path.getmtime(file)
            self._module_file = file
            self._func = self._module_co.func
            self._func_init = self._module_co.func_init
            self._method_init = self._module_co.method_init
            self._menu = self._module_co.menu
            for mname in self._module_co.method:
                m = getattr(self._module_co, mname)
                object(TreeDict, self).__setattr__(self, mname,
                                                   m.__get__(self, self, self.__class__))
        except Exception:
            logging.exception("Module Loading Failed")

    def load_classfile(self, file):
        self.setvar("module file", file)
        ####

    def tree_viewer_menu(self):
     # return MenuString, Handler, MenuImage
        return self.menu + \
            [('+Module', None, None),
                ('Reload', self.onReload, None),
                ('Edit File', self.onEditModule, None),
                ('Select File', self.onSelectModuleFile, None),
                #               ('Import File', self.onImportScriptFile, None),
                ('!', None, None),
                ('---', None, None)] + \
            super(PyClass, self).tree_viewer_menu()

    def onEditModule(self, e):
        self._module.load_module(self.getvar("module file"))
        if wx.GetKeyState(wx.WXK_CONTROL):
            self._module.edit_module()
            return
        handler = self.get_root_parent().app.GetEventHandler()
        evt = ifigure.events.TreeDictEvent(
            ifigure.events.EditFile,
            wx.ID_ANY)
        evt.SetTreeDict(self)
        evt.file = self._module._module_file
        handler.ProcessEvent(evt)

    def onSelectModuleFile(self, e=None):
        open_dlg = wx.FileDialog(None, message="Select module",
                                 wildcard='*.py', style=wx.FD_OPEN)
        if open_dlg.ShowModal() != wx.ID_OK:
            open_dlg.Destroy()
            raise ValueError
        file = open_dlg.GetPath()
        self.load_module(file)
        open_dlg.Destroy()
